"""GraphQL types for RBAC permission management."""

from __future__ import annotations

import uuid
from collections.abc import Iterable
from enum import StrEnum
from typing import TYPE_CHECKING, Annotated, Any, Self, override

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.common.data.permission.types import (
    OperationType,
    RBACElementType,
)
from ai.backend.manager.api.gql.base import OrderDirection
from ai.backend.manager.api.gql.rbac.types.entity_node import EntityNode
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.data.permission.permission import PermissionData
from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.base.creator import Creator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.permission_controller.creators import PermissionCreatorSpec
from ai.backend.manager.repositories.permission_controller.options import (
    ScopedPermissionConditions,
    ScopedPermissionOrders,
)
from ai.backend.manager.repositories.permission_controller.updaters import PermissionUpdaterSpec
from ai.backend.manager.types import OptionalState

if TYPE_CHECKING:
    from ai.backend.manager.api.gql.rbac.types.role import RoleGQL

# ==================== Enums ====================


@strawberry.enum(
    name="RBACElementType",
    description="Added in 26.3.0. Unified RBAC element type for scope-entity relationships",
)
class RBACElementTypeGQL(StrEnum):
    # Scope hierarchy
    DOMAIN = "domain"
    PROJECT = "project"
    USER = "user"

    # Root-query-enabled entities (scoped)
    SESSION = "session"
    VFOLDER = "vfolder"
    DEPLOYMENT = "deployment"
    MODEL_DEPLOYMENT = "model_deployment"
    KEYPAIR = "keypair"
    NOTIFICATION_CHANNEL = "notification_channel"
    NETWORK = "network"
    RESOURCE_GROUP = "resource_group"
    CONTAINER_REGISTRY = "container_registry"
    STORAGE_HOST = "storage_host"
    AGENT = "agent"
    KERNEL = "kernel"
    ROUTING = "routing"
    IMAGE = "image"
    ARTIFACT = "artifact"
    ARTIFACT_REGISTRY = "artifact_registry"
    SESSION_TEMPLATE = "session_template"
    APP_CONFIG = "app_config"

    # Root-query-enabled entities (superadmin-only)
    RESOURCE_PRESET = "resource_preset"
    USER_RESOURCE_POLICY = "user_resource_policy"
    KEYPAIR_RESOURCE_POLICY = "keypair_resource_policy"
    PROJECT_RESOURCE_POLICY = "project_resource_policy"
    ROLE = "role"
    AUDIT_LOG = "audit_log"
    EVENT_LOG = "event_log"

    # Auto-only entities used in permissions
    NOTIFICATION_RULE = "notification_rule"

    # Entity-level scopes
    ARTIFACT_REVISION = "artifact_revision"

    @classmethod
    def from_element(cls, value: RBACElementType) -> RBACElementTypeGQL:
        return cls(value.value)

    def to_element(self) -> RBACElementType:
        return RBACElementType(self.value)


@strawberry.enum(name="OperationType", description="Added in 26.3.0. RBAC operation type")
class OperationTypeGQL(StrEnum):
    CREATE = "create"
    READ = "read"
    UPDATE = "update"
    SOFT_DELETE = "soft-delete"
    HARD_DELETE = "hard-delete"
    GRANT_ALL = "grant:all"
    GRANT_READ = "grant:read"
    GRANT_UPDATE = "grant:update"
    GRANT_SOFT_DELETE = "grant:soft-delete"
    GRANT_HARD_DELETE = "grant:hard-delete"

    @classmethod
    def from_internal(cls, value: OperationType) -> OperationTypeGQL:
        try:
            return cls(value.value)
        except ValueError:
            raise InvalidAPIParameters(
                extra_msg=f"{value.value!r} is not a valid OperationTypeGQL"
            ) from None

    def to_internal(self) -> OperationType:
        try:
            return OperationType(self.value)
        except ValueError:
            raise InvalidAPIParameters(
                extra_msg=f"{self.value!r} is not a valid OperationType"
            ) from None


@strawberry.enum(description="Added in 26.3.0. Permission ordering field")
class PermissionOrderField(StrEnum):
    ID = "id"
    ENTITY_TYPE = "entity_type"


# ==================== Node Types ====================


@strawberry.type(name="Permission", description="Added in 26.3.0. RBAC scoped permission")
class PermissionGQL(Node):
    id: NodeID[str]
    role_id: uuid.UUID
    scope_type: RBACElementTypeGQL
    scope_id: str
    entity_type: RBACElementTypeGQL
    operation: OperationTypeGQL

    @classmethod
    async def resolve_nodes(  # type: ignore[override]
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.permission_loader.load_many([
            uuid.UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @strawberry.field(description="The role this permission belongs to.")  # type: ignore[misc]
    async def role(
        self, info: Info[StrawberryGQLContext]
    ) -> (
        Annotated[
            RoleGQL,
            strawberry.lazy("ai.backend.manager.api.gql.rbac.types.role"),
        ]
        | None
    ):
        from ai.backend.manager.api.gql.rbac.types.role import RoleGQL

        data = await info.context.data_loaders.role_loader.load(self.role_id)
        if data is None:
            return None
        return RoleGQL.from_dataclass(data)

    @strawberry.field(  # type: ignore[misc]
        description="The scope this permission applies to."
    )
    async def scope(
        self,
        *,
        info: Info[StrawberryGQLContext],
    ) -> EntityNode | None:
        from ai.backend.manager.api.gql.artifact.types import ArtifactRevision
        from ai.backend.manager.api.gql.deployment.types.deployment import ModelDeployment
        from ai.backend.manager.api.gql.domain_v2.types.node import DomainV2GQL
        from ai.backend.manager.api.gql.project_v2.types.node import ProjectV2GQL
        from ai.backend.manager.api.gql.rbac.types.role import RoleGQL
        from ai.backend.manager.api.gql.resource_group.types import ResourceGroupGQL
        from ai.backend.manager.api.gql.user.types.node import UserV2GQL

        element_type = self.scope_type.to_element()
        data_loaders = info.context.data_loaders
        match element_type:
            case RBACElementType.USER:
                user_data = await data_loaders.user_loader.load(uuid.UUID(self.scope_id))
                if user_data is None:
                    return None
                return UserV2GQL.from_data(user_data)
            case RBACElementType.PROJECT:
                project_data = await data_loaders.project_loader.load(uuid.UUID(self.scope_id))
                if project_data is None:
                    return None
                return ProjectV2GQL.from_data(project_data)
            case RBACElementType.DOMAIN:
                domain_data = await data_loaders.domain_loader.load(self.scope_id)
                if domain_data is None:
                    return None
                return DomainV2GQL.from_data(domain_data)
            case RBACElementType.ROLE:
                role_data = await data_loaders.role_loader.load(uuid.UUID(self.scope_id))
                if role_data is None:
                    return None
                return RoleGQL.from_dataclass(role_data)
            case RBACElementType.RESOURCE_GROUP:
                rg_data = await data_loaders.resource_group_loader.load(self.scope_id)
                if rg_data is None:
                    return None
                return ResourceGroupGQL.from_dataclass(rg_data)
            case RBACElementType.MODEL_DEPLOYMENT:
                deploy_data = await data_loaders.deployment_loader.load(uuid.UUID(self.scope_id))
                if deploy_data is None:
                    return None
                return ModelDeployment.from_dataclass(deploy_data)
            case RBACElementType.ARTIFACT_REVISION:
                rev_data = await data_loaders.artifact_revision_loader.load(
                    uuid.UUID(self.scope_id)
                )
                if rev_data is None:
                    return None
                return ArtifactRevision.from_dataclass(rev_data)
            case (
                RBACElementType.SESSION
                | RBACElementType.VFOLDER
                | RBACElementType.DEPLOYMENT
                | RBACElementType.KEYPAIR
                | RBACElementType.NOTIFICATION_CHANNEL
                | RBACElementType.NETWORK
                | RBACElementType.CONTAINER_REGISTRY
                | RBACElementType.STORAGE_HOST
                | RBACElementType.IMAGE
                | RBACElementType.ARTIFACT
                | RBACElementType.ARTIFACT_REGISTRY
                | RBACElementType.SESSION_TEMPLATE
                | RBACElementType.APP_CONFIG
                | RBACElementType.RESOURCE_PRESET
                | RBACElementType.USER_RESOURCE_POLICY
                | RBACElementType.KEYPAIR_RESOURCE_POLICY
                | RBACElementType.PROJECT_RESOURCE_POLICY
                | RBACElementType.AUDIT_LOG
                | RBACElementType.EVENT_LOG
                | RBACElementType.NOTIFICATION_RULE
                | RBACElementType.AGENT
                | RBACElementType.KERNEL
                | RBACElementType.ROUTING
            ):
                return None

    @classmethod
    def from_dataclass(cls, data: PermissionData) -> Self:
        return cls(
            id=ID(str(data.id)),
            role_id=data.role_id,
            scope_type=RBACElementTypeGQL.from_element(data.scope_type.to_element()),
            scope_id=data.scope_id,
            entity_type=RBACElementTypeGQL.from_element(data.entity_type.to_element()),
            operation=OperationTypeGQL.from_internal(data.operation),
        )


# ==================== Filter Types ====================


@strawberry.input(description="Added in 26.3.0. Filter for scoped permissions")
class PermissionFilter(GQLFilter):
    role_id: uuid.UUID | None = None
    scope_type: RBACElementTypeGQL | None = None
    entity_type: RBACElementTypeGQL | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []

        if self.role_id is not None:
            conditions.append(ScopedPermissionConditions.by_role_id(self.role_id))

        if self.scope_type is not None:
            conditions.append(
                ScopedPermissionConditions.by_scope_type(
                    self.scope_type.to_element().to_scope_type()
                )
            )

        if self.entity_type is not None:
            conditions.append(
                ScopedPermissionConditions.by_entity_type(
                    self.entity_type.to_element().to_entity_type()
                )
            )

        return conditions


# ==================== OrderBy Types ====================


@strawberry.input(description="Added in 26.3.0. Order by specification for permissions")
class PermissionOrderBy(GQLOrderBy):
    field: PermissionOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case PermissionOrderField.ID:
                return ScopedPermissionOrders.id(ascending)
            case PermissionOrderField.ENTITY_TYPE:
                return ScopedPermissionOrders.entity_type(ascending)


# ==================== Input Types ====================


@strawberry.input(description="Added in 26.3.0. Input for creating a scoped permission")
class CreatePermissionInput:
    role_id: uuid.UUID
    scope_type: RBACElementTypeGQL
    scope_id: str
    entity_type: RBACElementTypeGQL
    operation: OperationTypeGQL

    def to_creator(self) -> Creator[PermissionRow]:
        return Creator(
            spec=PermissionCreatorSpec(
                role_id=self.role_id,
                scope_type=self.scope_type.to_element().to_scope_type(),
                scope_id=self.scope_id,
                entity_type=self.entity_type.to_element().to_entity_type(),
                operation=self.operation.to_internal(),
            )
        )


@strawberry.input(description="Added in 26.3.0. Input for updating a scoped permission")
class UpdatePermissionInput:
    id: uuid.UUID
    scope_type: RBACElementTypeGQL | None = None
    scope_id: str | None = None
    entity_type: RBACElementTypeGQL | None = None
    operation: OperationTypeGQL | None = None

    def to_updater(self) -> Updater[PermissionRow]:
        spec = PermissionUpdaterSpec(
            scope_type=(
                OptionalState.update(self.scope_type.to_element().to_scope_type())
                if self.scope_type is not None
                else OptionalState.nop()
            ),
            scope_id=(
                OptionalState.update(self.scope_id)
                if self.scope_id is not None
                else OptionalState.nop()
            ),
            entity_type=(
                OptionalState.update(self.entity_type.to_element().to_entity_type())
                if self.entity_type is not None
                else OptionalState.nop()
            ),
            operation=(
                OptionalState.update(self.operation.to_internal())
                if self.operation is not None
                else OptionalState.nop()
            ),
        )
        return Updater(spec=spec, pk_value=self.id)


@strawberry.input(description="Added in 26.3.0. Input for deleting a scoped permission")
class DeletePermissionInput:
    id: uuid.UUID


# ==================== Payload Types ====================


@strawberry.type(description="Added in 26.3.0. Payload for delete permission mutation")
class DeletePermissionPayload:
    id: ID


# ==================== Connection Types ====================


@strawberry.type(
    name="ScopeEntityCombination",
    description="Added in 26.3.0. Valid scope-entity type combination for RBAC permissions.",
)
class ScopeEntityCombinationGQL:
    scope_type: RBACElementTypeGQL
    valid_entity_types: list[RBACElementTypeGQL]


PermissionEdge = Edge[PermissionGQL]


@strawberry.type(description="Added in 26.3.0. Permission connection")
class PermissionConnection(Connection[PermissionGQL]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.types import AcceleratorMetadata
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType

from .base import EtcdConfigAction


@dataclass
class GetResourceMetadataAction(EtcdConfigAction):
    """Action to get resource metadata with optional scaling group filter."""

    sgroup: str | None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET

    @override
    def entity_id(self) -> str | None:
        return self.sgroup


@dataclass
class GetResourceMetadataActionResult(BaseActionResult):
    """Result of getting resource metadata."""

    metadata: dict[str, AcceleratorMetadata]

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from ai.backend.common.clients.http_client import ClientPool, tcp_client_session_factory
from ai.backend.common.clients.prometheus.client import PrometheusClient
from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.manager.config.unified import ManagerUnifiedConfig


class PrometheusClientDependency(
    NonMonitorableDependencyProvider[ManagerUnifiedConfig, PrometheusClient],
):
    """Provides PrometheusClient with managed ClientPool lifecycle."""

    @property
    def stage_name(self) -> str:
        return "prometheus-client"

    @asynccontextmanager
    async def provide(self, setup_input: ManagerUnifiedConfig) -> AsyncIterator[PrometheusClient]:
        client_pool = ClientPool(tcp_client_session_factory)
        client = PrometheusClient(
            endpoint=f"http://{setup_input.metric.address.to_legacy()}/api/v1/",
            client_pool=client_pool,
        )
        try:
            yield client
        finally:
            await client_pool.close()



"""Export module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required, superadmin_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ExportHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_export_routes(handler: ExportHandler, route_deps: RouteDeps) -> RouteRegistry:
    """Build the export sub-application."""

    reg = RouteRegistry.create("export", route_deps.cors_options)

    # Report metadata endpoints
    reg.add(
        "GET",
        "/reports",
        handler.list_reports,
        middlewares=[auth_required, superadmin_required],
    )
    reg.add(
        "GET",
        "/reports/{report_key}",
        handler.get_report,
        middlewares=[auth_required, superadmin_required],
    )

    # CSV export endpoints
    reg.add(
        "POST",
        "/users/csv",
        handler.export_users_csv,
        middlewares=[auth_required, superadmin_required],
    )
    reg.add(
        "POST",
        "/sessions/csv",
        handler.export_sessions_csv,
        middlewares=[auth_required, superadmin_required],
    )
    reg.add(
        "POST",
        "/projects/csv",
        handler.export_projects_csv,
        middlewares=[auth_required, superadmin_required],
    )
    reg.add(
        "POST",
        "/keypairs/csv",
        handler.export_keypairs_csv,
        middlewares=[auth_required, superadmin_required],
    )
    reg.add(
        "POST",
        "/audit-logs/csv",
        handler.export_audit_logs_csv,
        middlewares=[auth_required, superadmin_required],
    )
    return reg



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.stream.repository import StreamRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class StreamRepositories:
    repository: StreamRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = StreamRepository(args.db)
        return cls(
            repository=repository,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience

from .db_source import ManagerAdminDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ManagerAdminRepository",)

manager_admin_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.MANAGER_ADMIN_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class ManagerAdminRepository:
    _db_source: ManagerAdminDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ManagerAdminDBSource(db)

    @manager_admin_repository_resilience.apply()
    async def count_active_sessions(self) -> int:
        """Count active sessions."""
        return await self._db_source.count_active_sessions()

    @manager_admin_repository_resilience.apply()
    async def update_agent_schedulable(self, agent_ids: list[str], schedulable: bool) -> int:
        """Update agent schedulable flags. Returns rows updated."""
        return await self._db_source.update_agent_schedulable(agent_ids, schedulable)



"""Manager API module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import superadmin_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ManagerHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_manager_api_routes(handler: ManagerHandler, route_deps: RouteDeps) -> RouteRegistry:
    """Build the manager sub-application."""

    reg = RouteRegistry.create("manager", route_deps.cors_options)

    # Public endpoints (no auth required)
    reg.add("GET", "/status", handler.fetch_manager_status)
    reg.add("GET", "/announcement", handler.get_announcement)
    reg.add("GET", "/prom", handler.get_manager_status_for_prom)

    # Superadmin endpoints
    reg.add("PUT", "/status", handler.update_manager_status, middlewares=[superadmin_required])
    reg.add(
        "POST",
        "/announcement",
        handler.update_announcement,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/scheduler/operation",
        handler.perform_scheduler_ops,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/scheduler/trigger",
        handler.scheduler_trigger,
        middlewares=[superadmin_required],
    )
    reg.add(
        "GET",
        "/scheduler/status",
        handler.scheduler_healthcheck,
        middlewares=[superadmin_required],
    )
    return reg



"""CreatorSpec implementations for keypair repository."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.manager.data.keypair.types import GeneratedKeyPairData, KeyPairCreator
from ai.backend.manager.models.keypair import KeyPairRow
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class KeyPairCreatorSpec(CreatorSpec[KeyPairRow]):
    """CreatorSpec for keypair creation with RBAC support."""

    creator: KeyPairCreator
    generated_data: GeneratedKeyPairData
    user_id: uuid.UUID
    email: str

    @override
    def build_row(self) -> KeyPairRow:
        return KeyPairRow.from_creator(
            self.creator,
            self.generated_data,
            self.user_id,
            self.email,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.dotfile.types import DotfileEntityKey, DotfileEntry, DotfileScope

from .base import DotfileAction


@dataclass
class ListOrGetDotfilesAction(DotfileAction):
    """Action to list all dotfiles or get a specific one by path."""

    scope: DotfileScope
    entity_key: DotfileEntityKey
    path: str | None = None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH

    @override
    def entity_id(self) -> str | None:
        return str(self.entity_key)


@dataclass
class ListOrGetDotfilesActionResult(BaseActionResult):
    """Result of listing or getting dotfiles.

    If a single dotfile was requested (path specified), ``entries`` contains
    exactly one element.
    """

    entries: list[DotfileEntry]

    @override
    def entity_id(self) -> str | None:
        return None



"""Resource Slot Type handler class using constructor dependency injection."""

from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING, Final

from ai.backend.common.api_handlers import APIResponse, BodyParam, PathParam
from ai.backend.common.dto.manager.resource_slot.request import (
    ResourceSlotTypePathParam,
    SearchResourceSlotTypesRequest,
)
from ai.backend.common.dto.manager.resource_slot.response import (
    GetResourceSlotTypeResponse,
    PaginationInfo,
    SearchResourceSlotTypesResponse,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.services.resource_slot.actions.get_resource_slot_type import (
    GetResourceSlotTypeAction,
)
from ai.backend.manager.services.resource_slot.actions.search_resource_slot_types import (
    SearchResourceSlotTypesAction,
)

from .adapter import ResourceSlotAdapter

if TYPE_CHECKING:
    from ai.backend.manager.services.resource_slot.processors import ResourceSlotProcessors

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ResourceSlotHandler:
    """Resource Slot Type API handler with constructor-injected dependencies."""

    def __init__(self, *, resource_slot: ResourceSlotProcessors) -> None:
        self._resource_slot = resource_slot
        self._adapter = ResourceSlotAdapter()

    async def search_resource_slot_types(
        self,
        body: BodyParam[SearchResourceSlotTypesRequest],
    ) -> APIResponse:
        """Search resource slot types with filters, orders, and pagination."""
        log.info("SEARCH_RESOURCE_SLOT_TYPES")

        querier = self._adapter.build_querier(body.parsed)

        action_result = await self._resource_slot.search_resource_slot_types.wait_for_complete(
            SearchResourceSlotTypesAction(querier=querier)
        )

        resp = SearchResourceSlotTypesResponse(
            items=[self._adapter.convert_to_dto(item) for item in action_result.items],
            pagination=PaginationInfo(
                total=action_result.total_count,
                offset=body.parsed.offset,
                limit=body.parsed.limit,
            ),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)

    async def get_resource_slot_type(
        self,
        path: PathParam[ResourceSlotTypePathParam],
    ) -> APIResponse:
        """Get a single resource slot type by slot_name."""
        slot_name = path.parsed.slot_name
        log.info("GET_RESOURCE_SLOT_TYPE (slot_name:{})", slot_name)

        action_result = await self._resource_slot.get_resource_slot_type.wait_for_complete(
            GetResourceSlotTypeAction(slot_name=slot_name)
        )

        resp = GetResourceSlotTypeResponse(
            item=self._adapter.convert_to_dto(action_result.item),
        )
        return APIResponse.build(status_code=HTTPStatus.OK, response_model=resp)



"""Artifact registry module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ArtifactRegistryHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_artifact_registry_routes(
    handler: ArtifactRegistryHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the artifact registry sub-application."""
    reg = RouteRegistry.create("artifact-registries", route_deps.cors_options)

    reg.add(
        "POST",
        "/scan",
        handler.scan_artifacts,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/delegation/scan",
        handler.delegate_scan_artifacts,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/delegation/import",
        handler.delegate_import_artifacts,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/search",
        handler.search_artifacts,
        middlewares=[auth_required],
    )
    reg.add(
        "GET",
        "/model/{model_id}",
        handler.scan_single_model,
        middlewares=[auth_required],
    )
    reg.add(
        "POST",
        "/models/batch",
        handler.scan_models,
        middlewares=[auth_required],
    )
    return reg



"""Permission fetcher functions."""

from __future__ import annotations

from functools import lru_cache

import strawberry
from strawberry import Info

from ai.backend.manager.api.gql.adapter import PaginationOptions, PaginationSpec
from ai.backend.manager.api.gql.base import encode_cursor
from ai.backend.manager.api.gql.rbac.types import (
    PermissionConnection,
    PermissionFilter,
    PermissionGQL,
    PermissionOrderBy,
)
from ai.backend.manager.api.gql.rbac.types.permission import PermissionEdge
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.repositories.base import QueryCondition
from ai.backend.manager.repositories.permission_controller.options import (
    ScopedPermissionConditions,
    ScopedPermissionOrders,
)
from ai.backend.manager.services.permission_contoller.actions.search_permissions import (
    SearchPermissionsAction,
)


@lru_cache(maxsize=1)
def get_permission_pagination_spec() -> PaginationSpec:
    return PaginationSpec(
        forward_order=ScopedPermissionOrders.id(ascending=False),
        backward_order=ScopedPermissionOrders.id(ascending=True),
        forward_condition_factory=ScopedPermissionConditions.by_cursor_forward,
        backward_condition_factory=ScopedPermissionConditions.by_cursor_backward,
        tiebreaker_order=PermissionRow.id.asc(),
    )


async def fetch_permissions(
    info: Info[StrawberryGQLContext],
    filter: PermissionFilter | None = None,
    order_by: list[PermissionOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
    base_conditions: list[QueryCondition] | None = None,
) -> PermissionConnection:
    querier = info.context.gql_adapter.build_querier(
        PaginationOptions(
            first=first,
            after=after,
            last=last,
            before=before,
            limit=limit,
            offset=offset,
        ),
        get_permission_pagination_spec(),
        filter=filter,
        order_by=order_by,
        base_conditions=base_conditions,
    )

    action_result = (
        await info.context.processors.permission_controller.search_permissions.wait_for_complete(
            SearchPermissionsAction(querier=querier)
        )
    )

    result = action_result.result
    edges = [
        PermissionEdge(
            node=PermissionGQL.from_dataclass(item),
            cursor=encode_cursor(str(item.id)),
        )
        for item in result.items
    ]
    return PermissionConnection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
        count=result.total_count,
    )



"""RBAC GQL resolvers package."""

from .entity import admin_entities
from .permission import (
    admin_create_permission,
    admin_delete_permission,
    admin_permissions,
    admin_update_permission,
    rbac_scope_entity_combinations,
)
from .role import (
    admin_assign_role,
    admin_bulk_assign_role,
    admin_bulk_revoke_role,
    admin_create_role,
    admin_delete_role,
    admin_purge_role,
    admin_revoke_role,
    admin_role,
    admin_role_assignments,
    admin_roles,
    admin_update_role,
    my_roles,
)

__all__ = [
    # Permission queries
    "admin_permissions",
    "rbac_scope_entity_combinations",
    # Entity queries
    "admin_entities",
    # Permission mutations
    "admin_create_permission",
    "admin_update_permission",
    "admin_delete_permission",
    # Role queries
    "admin_role",
    "admin_roles",
    "admin_role_assignments",
    "my_roles",
    # Role mutations
    "admin_create_role",
    "admin_update_role",
    "admin_delete_role",
    "admin_purge_role",
    "admin_assign_role",
    "admin_revoke_role",
    "admin_bulk_assign_role",
    "admin_bulk_revoke_role",
]



"""GraphQL types for resource group."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from decimal import Decimal
from enum import StrEnum
from typing import Self, assert_never, override

import strawberry
from strawberry import Info
from strawberry.relay import Node, NodeID

from ai.backend.common.types import PreemptionMode, PreemptionOrder, ResourceSlot
from ai.backend.manager.api.gql.base import OrderDirection, StringFilter
from ai.backend.manager.api.gql.fair_share.types.common import (
    ResourceSlotGQL,
    ResourceWeightEntryGQL,
    ResourceWeightEntryInputGQL,
)
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.api.gql.utils import dedent_strip
from ai.backend.manager.data.scaling_group.types import (
    PreemptionConfig as DataPreemptionConfig,
)
from ai.backend.manager.data.scaling_group.types import (
    ResourceInfo,
    ScalingGroupData,
    SchedulerType,
)
from ai.backend.manager.models.scaling_group.types import FairShareScalingGroupSpec
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.scaling_group.options import (
    ScalingGroupConditions,
    ScalingGroupOrders,
)
from ai.backend.manager.services.scaling_group.actions.get_resource_info import (
    GetResourceInfoAction,
)

__all__ = (
    "FairShareScalingGroupSpecGQL",
    "PreemptionConfigGQL",
    "PreemptionConfigInput",
    "PreemptionModeGQL",
    "PreemptionOrderGQL",
    "ResourceGroupFilterGQL",
    "ResourceGroupGQL",
    "ResourceGroupMetadataGQL",
    "ResourceGroupNetworkConfigGQL",
    "ResourceGroupOrderByGQL",
    "ResourceGroupOrderFieldGQL",
    "ResourceGroupSchedulerConfigGQL",
    "ResourceGroupStatusGQL",
    "ResourceInfoGQL",
    "SchedulerTypeGQL",
    "UpdateResourceGroupFairShareSpecInput",
    "UpdateResourceGroupFairShareSpecPayload",
    "UpdateResourceGroupInput",
    "UpdateResourceGroupPayload",
)


@strawberry.enum(
    name="SchedulerType",
    description="Added in 26.2.0. Type of scheduler used for session scheduling in a resource group.",
)
class SchedulerTypeGQL(StrEnum):
    """Scheduler type enumeration for GraphQL."""

    FIFO = "fifo"
    LIFO = "lifo"
    DRF = "drf"
    FAIR_SHARE = "fair-share"

    @classmethod
    def from_scheduler_type(cls, scheduler_type: SchedulerType) -> SchedulerTypeGQL:
        """Convert from data layer SchedulerType to GQL type."""
        match scheduler_type:
            case SchedulerType.FIFO:
                return cls.FIFO
            case SchedulerType.LIFO:
                return cls.LIFO
            case SchedulerType.DRF:
                return cls.DRF
            case SchedulerType.FAIR_SHARE:
                return cls.FAIR_SHARE


@strawberry.enum(
    name="PreemptionMode",
    description="Added in 26.3.0. How to preempt a session when preemption is triggered.",
)
class PreemptionModeGQL(StrEnum):
    """Preemption mode enumeration for GraphQL."""

    TERMINATE = "terminate"
    RESCHEDULE = "reschedule"

    @classmethod
    def from_preemption_mode(cls, mode: PreemptionMode) -> PreemptionModeGQL:
        """Convert from internal PreemptionMode to GQL type."""
        match mode:
            case PreemptionMode.TERMINATE:
                return cls.TERMINATE
            case PreemptionMode.RESCHEDULE:
                return cls.RESCHEDULE
            case _:
                assert_never(mode)


@strawberry.enum(
    name="PreemptionOrder",
    description="Added in 26.3.0. Tie-breaking order for same-priority sessions during preemption.",
)
class PreemptionOrderGQL(StrEnum):
    """Preemption order enumeration for GraphQL."""

    OLDEST = "oldest"
    NEWEST = "newest"

    @classmethod
    def from_preemption_order(cls, order: PreemptionOrder) -> PreemptionOrderGQL:
        """Convert from internal PreemptionOrder to GQL type."""
        match order:
            case PreemptionOrder.OLDEST:
                return cls.OLDEST
            case PreemptionOrder.NEWEST:
                return cls.NEWEST
            case _:
                assert_never(order)


@strawberry.type(
    name="PreemptionConfig",
    description="Added in 26.3.0. Preemption configuration for a resource group.",
)
class PreemptionConfigGQL:
    """Preemption configuration for GraphQL."""

    preemptible_priority: int = strawberry.field(
        description=(
            "Sessions with priority <= this value are eligible for preemption. Default is 5."
        )
    )
    order: PreemptionOrderGQL = strawberry.field(
        description=(
            "Tie-breaking order for sessions with the same priority during preemption. "
            "'oldest' preempts the oldest session first, 'newest' preempts the newest. "
            "Default is 'oldest'."
        )
    )
    mode: PreemptionModeGQL = strawberry.field(
        description=(
            "How to preempt a session when preemption is triggered. "
            "'terminate' destroys the session, 'reschedule' queues it for rescheduling. "
            "Default is 'terminate'."
        )
    )

    @classmethod
    def from_dataclass(cls, data: DataPreemptionConfig) -> Self:
        """Convert from data layer PreemptionConfig to GQL type."""
        return cls(
            preemptible_priority=data.preemptible_priority,
            order=PreemptionOrderGQL.from_preemption_order(data.order),
            mode=PreemptionModeGQL.from_preemption_mode(data.mode),
        )


@strawberry.type(
    name="ResourceGroupStatus",
    description="Added in 26.2.0. Status information for a resource group.",
)
class ResourceGroupStatusGQL:
    """Status information for a resource group."""

    is_active: bool = strawberry.field(
        description="Whether the resource group is active and can accept new sessions."
    )
    is_public: bool = strawberry.field(
        description="Whether the resource group is publicly accessible to all users."
    )


@strawberry.type(
    name="ResourceGroupMetadata",
    description="Added in 26.2.0. Metadata for a resource group.",
)
class ResourceGroupMetadataGQL:
    """Metadata for a resource group."""

    description: str | None = strawberry.field(
        description="Human-readable description of the resource group."
    )
    created_at: datetime = strawberry.field(
        description="Timestamp when the resource group was created."
    )


@strawberry.type(
    name="ResourceGroupNetworkConfig",
    description="Added in 26.2.0. Network configuration for a resource group.",
)
class ResourceGroupNetworkConfigGQL:
    """Network configuration for a resource group."""

    wsproxy_addr: str | None = strawberry.field(
        description="WebSocket proxy address for this resource group."
    )
    use_host_network: bool = strawberry.field(
        description="Whether to use host network mode for containers in this resource group."
    )


@strawberry.type(
    name="ResourceGroupSchedulerConfig",
    description="Added in 26.2.0. Scheduler configuration for a resource group.",
)
class ResourceGroupSchedulerConfigGQL:
    """Scheduler configuration for a resource group."""

    type: SchedulerTypeGQL = strawberry.field(
        description="Type of scheduler used for session scheduling (fifo, lifo, drf, fair-share)."
    )
    preemption: PreemptionConfigGQL = strawberry.field(
        description="Added in 26.3.0. Preemption configuration for this resource group."
    )


@strawberry.type(
    name="FairShareScalingGroupSpec",
    description=(
        "Added in 26.1.0. Fair share calculation configuration for a resource group. "
        "Defines parameters for computing fair share factors across domains, projects, and users."
    ),
)
class FairShareScalingGroupSpecGQL:
    """Fair share configuration for a resource group."""

    half_life_days: int = strawberry.field(
        description=(
            "Half-life for exponential decay in days. "
            "Determines how quickly historical usage loses significance. "
            "Default is 7 days."
        )
    )
    lookback_days: int = strawberry.field(
        description=(
            "Total lookback period in days for usage aggregation. "
            "Only usage within this window is considered. "
            "Default is 28 days."
        )
    )
    decay_unit_days: int = strawberry.field(
        description=(
            "Granularity of decay buckets in days. "
            "Usage is aggregated into buckets of this size. "
            "Default is 1 day."
        )
    )
    default_weight: Decimal = strawberry.field(
        description=(
            "Default weight for entities without explicit weight in this resource group. "
            "Default is 1.0."
        )
    )
    resource_weights: list[ResourceWeightEntryGQL] = strawberry.field(
        description=(
            "Added in 26.2.0. Weights for each resource type with default indicators. "
            "All resource types from capacity are included. "
            "Shows which resources use explicit vs default weights."
        )
    )

    @classmethod
    def from_model(
        cls,
        spec: FairShareScalingGroupSpec,
        uses_default_resources: frozenset[str],
    ) -> Self:
        """Convert from Pydantic model to GQL type.

        Args:
            spec: FairShareScalingGroupSpec with merged resource_weights
            uses_default_resources: Set of resource types using default weight
        """
        # Convert ResourceSlot to list[ResourceWeightEntryGQL]
        weight_entries = [
            ResourceWeightEntryGQL(
                resource_type=resource_type,
                weight=weight,
                uses_default=resource_type in uses_default_resources,
            )
            for resource_type, weight in spec.resource_weights.items()
        ]

        return cls(
            half_life_days=spec.half_life_days,
            lookback_days=spec.lookback_days,
            decay_unit_days=spec.decay_unit_days,
            default_weight=spec.default_weight,
            resource_weights=weight_entries,
        )


@strawberry.type(
    name="ResourceInfo",
    description=(
        "Added in 26.1.0. Resource information for a resource group. "
        "Provides aggregated resource metrics including capacity, used, and free resources."
    ),
)
class ResourceInfoGQL:
    """Resource information containing capacity, used, and free resource metrics."""

    capacity: ResourceSlotGQL = strawberry.field(
        description=(
            "Total available resources from ALIVE, schedulable agents in this resource group."
        )
    )
    used: ResourceSlotGQL = strawberry.field(
        description=(
            "Currently occupied resources from active kernels (RUNNING/TERMINATING status)."
        )
    )
    free: ResourceSlotGQL = strawberry.field(description="Available resources (capacity - used).")

    @classmethod
    def from_resource_info(cls, info: ResourceInfo) -> Self:
        """Convert from ResourceInfo dataclass to GQL type."""
        return cls(
            capacity=ResourceSlotGQL.from_slot_quantities(info.capacity),
            used=ResourceSlotGQL.from_slot_quantities(info.used),
            free=ResourceSlotGQL.from_slot_quantities(info.free),
        )


@strawberry.type(
    name="ResourceGroup",
    description="Added in 26.1.0. Resource group with structured configuration",
)
class ResourceGroupGQL(Node):
    id: NodeID[str] = strawberry.field(
        description="Relay-style global node identifier for the resource group"
    )
    name: str = strawberry.field(
        description=dedent_strip("""
            Unique name identifying the resource group.
            Used as primary key and referenced by agents, sessions, and resource presets.
        """)
    )
    status: ResourceGroupStatusGQL = strawberry.field(
        description="Added in 26.2.0. Status information including active and public flags."
    )
    metadata: ResourceGroupMetadataGQL = strawberry.field(
        description="Added in 26.2.0. Metadata including description and creation timestamp."
    )
    network: ResourceGroupNetworkConfigGQL = strawberry.field(
        description="Added in 26.2.0. Network configuration for the resource group."
    )
    scheduler: ResourceGroupSchedulerConfigGQL = strawberry.field(
        description=(
            "Added in 26.2.0. Scheduler configuration for the resource group. "
            "Use scheduler.type to check if fair-share scheduling is enabled."
        )
    )

    # Private field to store original fair share spec for lazy loading
    _fair_share_spec_data: strawberry.Private[FairShareScalingGroupSpec]

    @classmethod
    async def resolve_nodes(  # type: ignore[override]  # Strawberry Node uses AwaitableOrValue overloads incompatible with async def
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.resource_group_loader.load_many(node_ids)
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: ScalingGroupData) -> Self:
        return cls(
            id=data.name,
            name=data.name,
            status=ResourceGroupStatusGQL(
                is_active=data.status.is_active,
                is_public=data.status.is_public,
            ),
            metadata=ResourceGroupMetadataGQL(
                description=data.metadata.description if data.metadata.description else None,
                created_at=data.metadata.created_at,
            ),
            network=ResourceGroupNetworkConfigGQL(
                wsproxy_addr=data.network.wsproxy_addr if data.network.wsproxy_addr else None,
                use_host_network=data.network.use_host_network,
            ),
            scheduler=ResourceGroupSchedulerConfigGQL(
                type=SchedulerTypeGQL.from_scheduler_type(data.scheduler.name),
                preemption=PreemptionConfigGQL.from_dataclass(data.scheduler.options.preemption),
            ),
            _fair_share_spec_data=data.fair_share_spec,
        )

    @strawberry.field(  # type: ignore[misc]
        description=(
            "Added in 26.1.0. Fair share calculation configuration for this resource group. "
            "Defines decay parameters and resource weights for fair share factor computation. "
            "Resource weights are merged with capacity and include default indicators."
        )
    )
    async def fair_share_spec(
        self, info: Info[StrawberryGQLContext, None]
    ) -> FairShareScalingGroupSpecGQL:
        """Get fair share spec with merged resource weights from capacity.

        This is a lazy-loaded field that merges the resource group's fair share spec
        with its current capacity to provide complete resource weight information.
        """
        from ai.backend.common.types import ResourceSlot

        ctx = info.context

        # Get capacity from resource info
        action = GetResourceInfoAction(scaling_group=self.name)
        result = await ctx.processors.scaling_group.get_resource_info.wait_for_complete(action)
        capacity = result.resource_info.capacity

        # Merge resource weights with capacity
        merged = {}
        uses_default = []

        capacity_keys = {sq.slot_name for sq in capacity}
        for resource_type in capacity_keys:
            if resource_type in self._fair_share_spec_data.resource_weights.data:
                merged[resource_type] = self._fair_share_spec_data.resource_weights.data[
                    resource_type
                ]
            else:
                merged[resource_type] = self._fair_share_spec_data.default_weight
                uses_default.append(resource_type)

        # Create merged spec
        merged_spec = FairShareScalingGroupSpec(
            half_life_days=self._fair_share_spec_data.half_life_days,
            lookback_days=self._fair_share_spec_data.lookback_days,
            decay_unit_days=self._fair_share_spec_data.decay_unit_days,
            default_weight=self._fair_share_spec_data.default_weight,
            resource_weights=ResourceSlot(merged),
        )

        return FairShareScalingGroupSpecGQL.from_model(merged_spec, frozenset(uses_default))

    @strawberry.field(  # type: ignore[misc]
        description=(
            "Added in 26.1.0. Resource usage information for this resource group. "
            "Provides aggregated metrics for capacity, used, and free resources. "
            "This is a lazy-loaded field that queries agent and kernel data on demand."
        )
    )
    async def resource_info(self, info: Info[StrawberryGQLContext, None]) -> ResourceInfoGQL:
        """Get resource information for this resource group."""
        ctx = info.context
        action = GetResourceInfoAction(scaling_group=self.name)
        result = await ctx.processors.scaling_group.get_resource_info.wait_for_complete(action)
        return ResourceInfoGQL.from_resource_info(result.resource_info)


# Filter and OrderBy types


@strawberry.enum(
    name="ResourceGroupOrderField",
    description="Added in 26.1.0. Fields available for ordering resource groups",
)
class ResourceGroupOrderFieldGQL(StrEnum):
    NAME = "name"
    CREATED_AT = "created_at"
    IS_ACTIVE = "is_active"


@strawberry.input(
    name="ResourceGroupFilter",
    description="Added in 26.1.0. Filter for resource groups",
)
class ResourceGroupFilterGQL(GQLFilter):
    name: StringFilter | None = None
    description: StringFilter | None = None
    is_active: bool | None = None
    is_public: bool | None = None

    AND: list[ResourceGroupFilterGQL] | None = None
    OR: list[ResourceGroupFilterGQL] | None = None
    NOT: list[ResourceGroupFilterGQL] | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns a list containing a single combined QueryCondition that represents
        all filters with proper logical operators applied.
        """
        # Collect direct field conditions (these will be combined with AND)
        field_conditions: list[QueryCondition] = []

        # Apply name filter
        if self.name:
            name_condition = self.name.build_query_condition(
                contains_factory=ScalingGroupConditions.by_name_contains,
                equals_factory=ScalingGroupConditions.by_name_equals,
                starts_with_factory=ScalingGroupConditions.by_name_starts_with,
                ends_with_factory=ScalingGroupConditions.by_name_ends_with,
            )
            if name_condition:
                field_conditions.append(name_condition)

        # Apply description filter
        if self.description:
            description_condition = self.description.build_query_condition(
                contains_factory=ScalingGroupConditions.by_description_contains,
                equals_factory=ScalingGroupConditions.by_description_equals,
                starts_with_factory=ScalingGroupConditions.by_description_starts_with,
                ends_with_factory=ScalingGroupConditions.by_description_ends_with,
            )
            if description_condition:
                field_conditions.append(description_condition)

        # Apply is_active filter
        if self.is_active is not None:
            field_conditions.append(ScalingGroupConditions.by_is_active(self.is_active))

        # Apply is_public filter
        if self.is_public is not None:
            field_conditions.append(ScalingGroupConditions.by_is_public(self.is_public))

        # Handle AND logical operator - these are implicitly ANDed with field conditions
        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(
    name="ResourceGroupOrderBy",
    description="Added in 26.1.0. Order by specification for resource groups",
)
class ResourceGroupOrderByGQL(GQLOrderBy):
    field: ResourceGroupOrderFieldGQL
    direction: OrderDirection = OrderDirection.ASC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case ResourceGroupOrderFieldGQL.NAME:
                return ScalingGroupOrders.name(ascending)
            case ResourceGroupOrderFieldGQL.CREATED_AT:
                return ScalingGroupOrders.created_at(ascending)
            case ResourceGroupOrderFieldGQL.IS_ACTIVE:
                return ScalingGroupOrders.is_active(ascending)


@strawberry.input(
    name="PreemptionConfigInput",
    description="Added in 26.3.0. Input for preemption configuration.",
)
class PreemptionConfigInput:
    """Input for preemption configuration. Replaces entire preemption config when provided."""

    preemptible_priority: int = strawberry.field(
        default=5,
        description=("Sessions with priority <= this value are preemptible. Default is 5."),
    )
    order: PreemptionOrderGQL = strawberry.field(
        default=PreemptionOrderGQL.OLDEST,
        description=(
            "Tie-breaking order for same-priority sessions (OLDEST, NEWEST). Default is OLDEST."
        ),
    )
    mode: PreemptionModeGQL = strawberry.field(
        default=PreemptionModeGQL.TERMINATE,
        description=("How to preempt sessions (TERMINATE, RESCHEDULE). Default is TERMINATE."),
    )


# Mutation Input/Payload types


@strawberry.input(
    name="UpdateResourceGroupFairShareSpecInput",
    description=(
        "Added in 26.1.0. Input for updating resource group fair share configuration. "
        "All fields are optional - only provided fields will be updated, others retain existing values."
    ),
)
class UpdateResourceGroupFairShareSpecInput:
    """Partial update input for fair share spec. All fields optional for partial update."""

    resource_group_name: str = strawberry.field(description="Name of the resource group to update.")
    half_life_days: int | None = strawberry.field(
        default=None,
        description="Half-life for exponential decay in days. Leave null to keep existing value.",
    )
    lookback_days: int | None = strawberry.field(
        default=None,
        description="Total lookback period in days. Leave null to keep existing value.",
    )
    decay_unit_days: int | None = strawberry.field(
        default=None,
        description="Granularity of decay buckets in days. Leave null to keep existing value.",
    )
    default_weight: Decimal | None = strawberry.field(
        default=None,
        description="Default weight for entities. Leave null to keep existing value.",
    )
    resource_weights: list[ResourceWeightEntryInputGQL] | None = strawberry.field(
        default=None,
        description=(
            "Resource weights for fair share calculation. "
            "Each entry specifies a resource type and its weight multiplier. "
            "Only provided resource types are updated (partial update). "
            "Set weight to null to remove that resource type (revert to default). "
            "Leave the entire list null to keep all existing values."
        ),
    )

    def merge_with(self, existing: FairShareScalingGroupSpec) -> FairShareScalingGroupSpec:
        """Merge partial input with existing spec, returning a new complete spec."""
        # Merge resource_weights: partial update with deletion support
        merged_resource_weights = existing.resource_weights
        if self.resource_weights is not None:
            # Start with existing weights
            merged_weights_dict = dict(existing.resource_weights)
            for entry in self.resource_weights:
                if entry.weight is None:
                    # Remove the resource type (revert to default)
                    merged_weights_dict.pop(entry.resource_type, None)
                else:
                    # Update or add the resource type
                    merged_weights_dict[entry.resource_type] = entry.weight
            merged_resource_weights = ResourceSlot(merged_weights_dict)

        return FairShareScalingGroupSpec(
            half_life_days=self.half_life_days
            if self.half_life_days is not None
            else existing.half_life_days,
            lookback_days=self.lookback_days
            if self.lookback_days is not None
            else existing.lookback_days,
            decay_unit_days=self.decay_unit_days
            if self.decay_unit_days is not None
            else existing.decay_unit_days,
            default_weight=self.default_weight
            if self.default_weight is not None
            else existing.default_weight,
            resource_weights=merged_resource_weights,
        )


@strawberry.type(
    name="UpdateResourceGroupFairShareSpecPayload",
    description="Added in 26.1.0. Payload for resource group fair share spec update mutation.",
)
class UpdateResourceGroupFairShareSpecPayload:
    """Payload for fair share spec update mutation."""

    resource_group: ResourceGroupGQL = strawberry.field(
        description="The updated resource group with new fair share configuration."
    )


@strawberry.input(
    name="UpdateResourceGroupInput",
    description=(
        "Added in 26.2.0. Resource group configuration update input. "
        "All fields are optional - only provided fields will be updated. "
        "Supports all ScalingGroupUpdaterSpec fields (except fair_share, use separate mutation)."
    ),
)
class UpdateResourceGroupInput:
    """Input for updating resource group configuration. All fields optional for partial update."""

    resource_group_name: str = strawberry.field(description="Name of the resource group to update.")

    # Status fields (ScalingGroupStatusUpdaterSpec)
    is_active: bool | None = strawberry.field(
        default=None,
        description="Whether the resource group is active. Leave null to keep existing value.",
    )
    is_public: bool | None = strawberry.field(
        default=None,
        description="Whether the resource group is public. Leave null to keep existing value.",
    )

    # Metadata fields (ScalingGroupMetadataUpdaterSpec)
    description: str | None = strawberry.field(
        default=None,
        description="Human-readable description. Leave null to keep existing value.",
    )

    # Network config fields (ScalingGroupNetworkConfigUpdaterSpec)
    app_proxy_addr: str | None = strawberry.field(
        default=None,
        description="App proxy address. Leave null to keep existing value.",
    )
    appproxy_api_token: str | None = strawberry.field(
        default=None,
        description="App proxy API token. Leave null to keep existing value.",
    )
    use_host_network: bool | None = strawberry.field(
        default=None,
        description="Whether to use host network mode. Leave null to keep existing value.",
    )

    # Scheduler config fields (ScalingGroupSchedulerConfigUpdaterSpec)
    scheduler_type: SchedulerTypeGQL | None = strawberry.field(
        default=None,
        description=(
            "Scheduler type (FIFO, LIFO, DRF, FAIR_SHARE). Leave null to keep existing value."
        ),
    )
    preemption: PreemptionConfigInput | None = strawberry.field(
        default=None,
        description=(
            "Added in 26.3.0. Preemption configuration. When provided, replaces the entire "
            "preemption config. Leave null to keep existing value."
        ),
    )


@strawberry.type(
    name="UpdateResourceGroupPayload",
    description="Added in 26.2.0. Payload for resource group update mutation.",
)
class UpdateResourceGroupPayload:
    """Payload for resource group update mutation."""

    resource_group: ResourceGroupGQL = strawberry.field(
        description="The updated resource group with new configuration."
    )



"""Scheduling history module registrar."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import superadmin_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import SchedulingHistoryHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_scheduling_history_routes(
    handler: SchedulingHistoryHandler, route_deps: RouteDeps
) -> RouteRegistry:
    """Build the scheduling history sub-application."""

    reg = RouteRegistry.create("scheduling-history", route_deps.cors_options)

    reg.add(
        "POST",
        "/sessions/search",
        handler.search_session_history,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/deployments/search",
        handler.search_deployment_history,
        middlewares=[superadmin_required],
    )
    reg.add(
        "POST",
        "/routes/search",
        handler.search_route_history,
        middlewares=[superadmin_required],
    )
    return reg



"""
Repository layer exceptions.
"""

from __future__ import annotations

from typing import Any

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)


class RepositoryError(BackendAIError):
    """Base class for repository layer errors."""

    error_type = "https://api.backend.ai/probs/repository-error"
    error_title = "Repository operation failed."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UpsertEmptyResultError(RepositoryError):
    """Raised when upsert operation returns no rows."""

    error_type = "https://api.backend.ai/probs/upsert-empty-result"
    error_title = "Upsert operation did not return any row."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.CREATE,
            error_detail=ErrorDetail.INTERNAL_ERROR,
        )


class UnsupportedCompositePrimaryKeyError(RepositoryError):
    """Raised when an operation requires a single-column primary key but the table has a composite key."""

    error_type = "https://api.backend.ai/probs/unsupported-composite-pk"
    error_title = "Unsupported composite primary key."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.ACCESS,
            error_detail=ErrorDetail.INVALID_PARAMETERS,
        )


class RepositoryIntegrityError(RepositoryError, web.HTTPConflict):
    """Base class for integrity constraint violation errors.

    Carries structured attributes extracted from asyncpg diagnostics.
    """

    error_type = "https://api.backend.ai/probs/integrity-error"
    error_title = "Integrity constraint violated."

    constraint_name: str | None
    table_name: str | None
    column_name: str | None
    detail: str | None
    pgcode: str | None

    def __init__(
        self,
        extra_msg: str | None = None,
        extra_data: Any | None = None,
        *,
        constraint_name: str | None = None,
        table_name: str | None = None,
        column_name: str | None = None,
        detail: str | None = None,
        pgcode: str | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(extra_msg=extra_msg, extra_data=extra_data, **kwargs)
        self.constraint_name = constraint_name
        self.table_name = table_name
        self.column_name = column_name
        self.detail = detail
        self.pgcode = pgcode

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.CONFLICT,
        )


class UniqueConstraintViolationError(RepositoryIntegrityError):
    """Raised when a unique constraint is violated (SQLSTATE 23505)."""

    error_type = "https://api.backend.ai/probs/unique-constraint-violation"
    error_title = "Unique constraint violated."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.ALREADY_EXISTS,
        )


class ForeignKeyViolationError(RepositoryIntegrityError):
    """Raised when a foreign key constraint is violated (SQLSTATE 23503)."""

    error_type = "https://api.backend.ai/probs/foreign-key-violation"
    error_title = "Foreign key constraint violated."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.NOT_FOUND,
        )


class CheckConstraintViolationError(RepositoryIntegrityError):
    """Raised when a check constraint is violated (SQLSTATE 23514)."""

    error_type = "https://api.backend.ai/probs/check-constraint-violation"
    error_title = "Check constraint violated."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.CONFLICT,
        )


class NotNullViolationError(RepositoryIntegrityError):
    """Raised when a not-null constraint is violated (SQLSTATE 23502).

    Uses HTTP 400 Bad Request instead of the default 409 Conflict.
    """

    error_type = "https://api.backend.ai/probs/not-null-violation"
    error_title = "Not-null constraint violated."
    status_code = 400

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.BAD_REQUEST,
        )


class ExclusionViolationError(RepositoryIntegrityError):
    """Raised when an exclusion constraint is violated (SQLSTATE 23P01)."""

    error_type = "https://api.backend.ai/probs/exclusion-violation"
    error_title = "Exclusion constraint violated."

    def error_code(self) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DATABASE,
            operation=ErrorOperation.GENERIC,
            error_detail=ErrorDetail.CONFLICT,
        )



from .create_cluster_template import (
    CreateClusterTemplateAction,
    CreateClusterTemplateActionResult,
)
from .create_task_template import (
    CreatedTaskTemplateItem,
    CreateTaskTemplateAction,
    CreateTaskTemplateActionResult,
    TaskTemplateItemInput,
)
from .delete_cluster_template import (
    DeleteClusterTemplateAction,
    DeleteClusterTemplateActionResult,
)
from .delete_task_template import DeleteTaskTemplateAction, DeleteTaskTemplateActionResult
from .get_cluster_template import GetClusterTemplateAction, GetClusterTemplateActionResult
from .get_task_template import GetTaskTemplateAction, GetTaskTemplateActionResult
from .list_cluster_templates import ListClusterTemplatesAction, ListClusterTemplatesActionResult
from .list_task_templates import ListTaskTemplatesAction, ListTaskTemplatesActionResult
from .update_cluster_template import UpdateClusterTemplateAction, UpdateClusterTemplateActionResult
from .update_task_template import UpdateTaskTemplateAction, UpdateTaskTemplateActionResult

__all__ = (
    "CreateClusterTemplateAction",
    "CreateClusterTemplateActionResult",
    "CreateTaskTemplateAction",
    "CreateTaskTemplateActionResult",
    "CreatedTaskTemplateItem",
    "DeleteClusterTemplateAction",
    "DeleteClusterTemplateActionResult",
    "DeleteTaskTemplateAction",
    "DeleteTaskTemplateActionResult",
    "GetClusterTemplateAction",
    "GetClusterTemplateActionResult",
    "GetTaskTemplateAction",
    "GetTaskTemplateActionResult",
    "ListClusterTemplatesAction",
    "ListClusterTemplatesActionResult",
    "ListTaskTemplatesAction",
    "ListTaskTemplatesActionResult",
    "TaskTemplateItemInput",
    "UpdateClusterTemplateAction",
    "UpdateClusterTemplateActionResult",
    "UpdateTaskTemplateAction",
    "UpdateTaskTemplateActionResult",
)



"""Error log module registrar.

Lifecycle management (GlobalTimer for log cleanup, event dispatcher
integration) is handled by the DependencyComposer:

* Event consumer: ``event_dispatcher.handlers.log_cleanup``
* GlobalTimer: ``dependencies.processing.log_cleanup_timer``
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.manager.api.rest.middleware.auth import auth_required
from ai.backend.manager.api.rest.routing import RouteRegistry

from .handler import ErrorLogHandler

if TYPE_CHECKING:
    from ai.backend.manager.api.rest.types import RouteDeps


def register_error_log_routes(handler: ErrorLogHandler, route_deps: RouteDeps) -> RouteRegistry:
    """Build the error log sub-application."""

    reg = RouteRegistry.create("logs", route_deps.cors_options)

    reg.add(
        "POST",
        "/error",
        handler.append,
        middlewares=[route_deps.read_status_mw, auth_required],
    )
    reg.add(
        "GET",
        "/error",
        handler.list_logs,
        middlewares=[auth_required, route_deps.read_status_mw],
    )
    reg.add(
        "POST",
        "/error/{log_id}/clear",
        handler.mark_cleared,
        middlewares=[auth_required, route_deps.read_status_mw],
    )
    return reg



"""CreatorSpec implementations for artifact entities."""

from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.manager.data.artifact.types import ArtifactType
from ai.backend.manager.models.artifact import ArtifactRow
from ai.backend.manager.repositories.base.creator import CreatorSpec


@dataclass
class ArtifactCreatorSpec(CreatorSpec[ArtifactRow]):
    """CreatorSpec for artifact creation."""

    name: str
    type: ArtifactType
    registry_id: uuid.UUID
    registry_type: ArtifactRegistryType | str
    source_registry_id: uuid.UUID
    source_registry_type: ArtifactRegistryType | str
    readonly: bool = True
    description: str | None = None
    extra: Any | None = None

    @override
    def build_row(self) -> ArtifactRow:
        return ArtifactRow(
            name=self.name,
            type=self.type,
            registry_id=self.registry_id,
            registry_type=self.registry_type,
            source_registry_id=self.source_registry_id,
            source_registry_type=self.source_registry_type,
            readonly=self.readonly,
            description=self.description,
            extra=self.extra,
        )



import datetime as dt
import logging
from datetime import UTC, datetime

import sqlalchemy as sa
from dateutil.relativedelta import relativedelta

from ai.backend.common import validators as tx
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.event_types.log.anycast import DoLogCleanupEvent
from ai.backend.common.types import AgentId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.models.error_logs import error_logs
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class LogCleanupEventHandler:
    _etcd: AsyncEtcd
    _db: ExtendedAsyncSAEngine

    def __init__(
        self,
        etcd: AsyncEtcd,
        db: ExtendedAsyncSAEngine,
    ) -> None:
        self._etcd = etcd
        self._db = db

    async def handle_log_cleanup(
        self,
        _context: None,
        _source: AgentId,
        _event: DoLogCleanupEvent,
    ) -> None:
        raw_lifetime = await self._etcd.get("config/logs/error/retention")
        if raw_lifetime is None:
            raw_lifetime = "90d"
        lifetime: dt.timedelta | relativedelta
        try:
            lifetime = tx.TimeDuration().check(raw_lifetime)
        except ValueError:
            lifetime = dt.timedelta(days=90)
            log.warning(
                "Failed to parse the error log retention period ({}) read from etcd; "
                "falling back to 90 days",
                raw_lifetime,
            )
        boundary = datetime.now(UTC) - lifetime
        async with self._db.begin() as conn:
            query = sa.delete(error_logs).where(error_logs.c.created_at < boundary)
            result = await conn.execute(query)
            if result.rowcount > 0:
                log.info("Cleaned up {} log(s) filed before {}", result.rowcount, boundary)



from __future__ import annotations

from .loader import (
    load_assignments_by_role_ids,
    load_element_associations_by_ids,
    load_entities_by_type_and_ids,
    load_permissions_by_ids,
    load_permissions_by_role_ids,
    load_role_assignments_by_ids,
    load_role_assignments_by_role_and_user_ids,
    load_roles_by_ids,
)

__all__ = [
    "load_assignments_by_role_ids",
    "load_element_associations_by_ids",
    "load_entities_by_type_and_ids",
    "load_roles_by_ids",
    "load_permissions_by_ids",
    "load_permissions_by_role_ids",
    "load_role_assignments_by_ids",
    "load_role_assignments_by_role_and_user_ids",
]



import uuid
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, override

from ai.backend.common.types import AccessKey, RuntimeVariant, VFolderMount
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.model_serving.types import ServiceConfig
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class ValidateModelServiceAction(ModelServiceAction):
    requester_access_key: AccessKey
    owner_access_key: AccessKey
    requester_uuid: uuid.UUID
    requester_role: UserRole
    requester_domain: str
    keypair_resource_policy: dict[str, Any]
    domain_name: str
    group_name: str
    config: ServiceConfig
    replicas: int
    runtime_variant: RuntimeVariant
    max_session_count_per_model_session: int
    owner_access_key_override: AccessKey | None

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.GET


@dataclass
class ValidateModelServiceActionResult(BaseActionResult):
    model_id: uuid.UUID
    model_definition_path: str | None
    requester_access_key: AccessKey
    owner_access_key: AccessKey
    owner_uuid: uuid.UUID
    owner_role: UserRole
    group_id: uuid.UUID
    resource_policy: dict[str, Any]
    scaling_group: str
    extra_mounts: Sequence[VFolderMount]

    @override
    def entity_id(self) -> str | None:
        return None



"""UpdaterSpec implementations for prometheus query preset repository."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, override

from ai.backend.manager.models.prometheus_query_preset.row import (
    PresetOptions,
    PrometheusQueryPresetRow,
)
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState, TriState


@dataclass
class PrometheusQueryPresetUpdaterSpec(UpdaterSpec[PrometheusQueryPresetRow]):
    """UpdaterSpec for prometheus query preset updates."""

    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    metric_name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    query_template: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    time_window: TriState[str] = field(default_factory=TriState[str].nop)
    filter_labels: OptionalState[list[str]] = field(default_factory=OptionalState[list[str]].nop)
    group_labels: OptionalState[list[str]] = field(default_factory=OptionalState[list[str]].nop)

    @property
    @override
    def row_class(self) -> type[PrometheusQueryPresetRow]:
        return PrometheusQueryPresetRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.metric_name.update_dict(to_update, "metric_name")
        self.query_template.update_dict(to_update, "query_template")
        self.time_window.update_dict(to_update, "time_window")
        # filter_labels and group_labels are stored together in the JSONB `options` column.
        # DBSource._merge_partial_options() ensures both fields are filled before this runs,
        # so the [] fallback is a safety net that should not be reached in normal flow.
        filter_value = self.filter_labels.optional_value()
        group_value = self.group_labels.optional_value()
        if filter_value is not None or group_value is not None:
            to_update["options"] = PresetOptions(
                filter_labels=filter_value if filter_value is not None else [],
                group_labels=group_value if group_value is not None else [],
            )
        return to_update



from __future__ import annotations

import logging
from collections.abc import Awaitable, Callable
from datetime import timedelta
from typing import Final

from ai.backend.common import validators as tx
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.types import AccessKey, KernelId, SessionId
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.idle import AppStreamingStatus, IdleCheckerHost
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.repositories.stream.repository import StreamRepository
from ai.backend.manager.services.stream.actions.execute_in_stream import (
    ExecuteInStreamAction,
    ExecuteInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.gc_stale_connections import (
    GCStaleConnectionsAction,
    GCStaleConnectionsActionResult,
)
from ai.backend.manager.services.stream.actions.get_streaming_session import (
    GetStreamingSessionAction,
    GetStreamingSessionActionResult,
)
from ai.backend.manager.services.stream.actions.interrupt_in_stream import (
    InterruptInStreamAction,
    InterruptInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.restart_in_stream import (
    RestartInStreamAction,
    RestartInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.start_service_in_stream import (
    StartServiceInStreamAction,
    StartServiceInStreamActionResult,
)
from ai.backend.manager.services.stream.actions.track_connection import (
    TrackConnectionAction,
    TrackConnectionActionResult,
)
from ai.backend.manager.services.stream.actions.untrack_connection import (
    UntrackConnectionAction,
    UntrackConnectionActionResult,
)

log: Final = BraceStyleAdapter(logging.getLogger(__spec__.name))


class StreamService:
    _repository: StreamRepository
    _registry: AgentRegistry
    _valkey_live: ValkeyLiveClient
    _idle_checker_host: IdleCheckerHost
    _etcd: AsyncEtcd

    def __init__(
        self,
        repository: StreamRepository,
        registry: AgentRegistry,
        valkey_live: ValkeyLiveClient,
        idle_checker_host: IdleCheckerHost,
        etcd: AsyncEtcd,
    ) -> None:
        self._repository = repository
        self._registry = registry
        self._valkey_live = valkey_live
        self._idle_checker_host = idle_checker_host
        self._etcd = etcd

    async def get_streaming_session(
        self, action: GetStreamingSessionAction
    ) -> GetStreamingSessionActionResult:
        session = await self._repository.get_streaming_session(
            action.session_name, AccessKey(action.access_key)
        )
        kernel = session.main_kernel
        return GetStreamingSessionActionResult(
            session_id=str(session.id),
            kernel_id=str(kernel.id),
            kernel_host=kernel.kernel_host,
            agent_addr=kernel.agent_addr,
            repl_in_port=kernel.repl_in_port,
            repl_out_port=kernel.repl_out_port,
            service_ports=kernel.service_ports or [],
        )

    async def execute_in_stream(self, action: ExecuteInStreamAction) -> ExecuteInStreamActionResult:
        session = await self._repository.get_streaming_session(
            action.session_name, AccessKey(action.access_key)
        )
        result = await self._registry.execute(
            session,
            action.api_version,
            action.run_id,
            action.mode,
            action.code,
            action.opts,
            flush_timeout=action.flush_timeout,
        )
        return ExecuteInStreamActionResult(result=dict(result))

    async def restart_in_stream(self, action: RestartInStreamAction) -> RestartInStreamActionResult:
        session = await self._repository.get_streaming_session(
            action.session_name, AccessKey(action.access_key)
        )
        await self._registry.restart_session(session)
        return RestartInStreamActionResult()

    async def interrupt_in_stream(
        self, action: InterruptInStreamAction
    ) -> InterruptInStreamActionResult:
        session = await self._repository.get_streaming_session(
            action.session_name, AccessKey(action.access_key)
        )
        result = await self._registry.interrupt_session(session)
        return InterruptInStreamActionResult(result=dict(result))

    async def start_service_in_stream(
        self, action: StartServiceInStreamAction
    ) -> StartServiceInStreamActionResult:
        session = await self._repository.get_streaming_session(
            action.session_name, AccessKey(action.access_key)
        )
        result = await self._registry.start_service(session, action.service, action.opts)
        return StartServiceInStreamActionResult(result=dict(result))

    def create_connection_refresh_callback(
        self,
        kernel_id: KernelId,
        service: str,
        stream_id: str,
    ) -> Callable[..., Awaitable[None]]:
        async def update_connection_tracker() -> None:
            await self._valkey_live.update_app_connection_tracker(
                str(kernel_id), service, stream_id
            )

        return update_connection_tracker

    async def track_connection(self, action: TrackConnectionAction) -> TrackConnectionActionResult:
        await self._valkey_live.update_connection_tracker(
            str(action.kernel_id), action.service, action.stream_id
        )
        await self._idle_checker_host.update_app_streaming_status(
            action.session_id,
            AppStreamingStatus.HAS_ACTIVE_CONNECTIONS,
        )
        return TrackConnectionActionResult(kernel_id=str(action.kernel_id))

    async def untrack_connection(
        self, action: UntrackConnectionAction
    ) -> UntrackConnectionActionResult:
        await self._valkey_live.remove_connection_tracker(
            str(action.kernel_id), action.service, action.stream_id
        )
        remaining_count = await self._valkey_live.count_active_connections(str(action.kernel_id))
        if remaining_count == 0:
            await self._idle_checker_host.update_app_streaming_status(
                action.session_id,
                AppStreamingStatus.NO_ACTIVE_CONNECTIONS,
            )
        return UntrackConnectionActionResult(
            kernel_id=str(action.kernel_id),
            remaining_count=remaining_count,
        )

    async def gc_stale_connections(
        self, action: GCStaleConnectionsAction
    ) -> GCStaleConnectionsActionResult:
        no_packet_timeout: timedelta = tx.TimeDuration().check(
            await self._etcd.get("config/idle/app-streaming-packet-timeout") or "5m",
        )
        now = await self._valkey_live.get_server_time()
        removed_sessions: list[str] = []
        for session_id in action.active_session_ids:
            prev_remaining = await self._valkey_live.count_active_connections(str(session_id))
            removed_count = await self._valkey_live.remove_stale_connections(
                str(session_id),
                now - no_packet_timeout.total_seconds(),
            )
            remaining = await self._valkey_live.count_active_connections(str(session_id))
            log.debug(
                "conn_tracker: gc {} removed/remaining = {}/{}",
                session_id,
                removed_count,
                remaining,
            )
            if prev_remaining > 0 and remaining == 0:
                await self._idle_checker_host.update_app_streaming_status(
                    SessionId(session_id),
                    AppStreamingStatus.NO_ACTIVE_CONNECTIONS,
                )
                removed_sessions.append(str(session_id))
        return GCStaleConnectionsActionResult(removed_sessions=removed_sessions)



from __future__ import annotations

from collections.abc import Sequence

from ai.backend.common.types import SessionId
from ai.backend.manager.data.session.types import SessionData
from ai.backend.manager.repositories.base import BatchQuerier, NoPagination
from ai.backend.manager.repositories.scheduler.options import SessionConditions
from ai.backend.manager.services.session.actions.search import SearchSessionsAction
from ai.backend.manager.services.session.processors import SessionProcessors


async def load_sessions_by_ids(
    processor: SessionProcessors,
    session_ids: Sequence[SessionId],
) -> list[SessionData | None]:
    if not session_ids:
        return []

    querier = BatchQuerier(
        pagination=NoPagination(),
        conditions=[SessionConditions.by_ids(session_ids)],
    )

    action_result = await processor.search_sessions.wait_for_complete(
        SearchSessionsAction(querier=querier)
    )

    session_map: dict[SessionId, SessionData] = {
        SessionId(data.id): data for data in action_result.data
    }
    return [session_map.get(session_id) for session_id in session_ids]



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from ai.backend.common.dependencies import NonMonitorableDependencyProvider
from ai.backend.manager.clients.appproxy.client import AppProxyClientPool


class AppProxyClientPoolDependency(
    NonMonitorableDependencyProvider[None, AppProxyClientPool],
):
    """Provides AppProxyClientPool lifecycle management."""

    @property
    def stage_name(self) -> str:
        return "appproxy-client-pool"

    @asynccontextmanager
    async def provide(self, setup_input: None) -> AsyncIterator[AppProxyClientPool]:
        """Initialize and provide an app proxy client pool.

        Args:
            setup_input: Not used (no dependencies required)

        Yields:
            Initialized AppProxyClientPool
        """
        pool = AppProxyClientPool()
        try:
            yield pool
        finally:
            await pool.close()



"""GraphQL types for RBAC role management."""

from __future__ import annotations

import uuid
from collections.abc import Iterable
from datetime import UTC, datetime
from enum import StrEnum
from typing import TYPE_CHECKING, Annotated, Any, Self, override

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.common.data.permission.types import (
    RoleSource,
    RoleStatus,
)
from ai.backend.manager.api.gql.base import OrderDirection, StringFilter
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.data.permission.role import (
    AssignedUserData,
    BulkUserRoleRevocationInput,
    RoleData,
    RoleDetailData,
    UserRoleAssignmentData,
    UserRoleAssignmentInput,
    UserRoleRevocationData,
    UserRoleRevocationInput,
)
from ai.backend.manager.models.rbac_models.role import RoleRow
from ai.backend.manager.models.rbac_models.user_role import UserRoleRow
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.base.creator import BulkCreator, Creator
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.repositories.permission_controller.creators import (
    RoleCreatorSpec,
    UserRoleCreatorSpec,
)
from ai.backend.manager.repositories.permission_controller.options import (
    AssignedUserConditions,
    AssignedUserOrders,
    RoleConditions,
    RoleOrders,
)
from ai.backend.manager.repositories.permission_controller.updaters import RoleUpdaterSpec
from ai.backend.manager.types import OptionalState, TriState

if TYPE_CHECKING:
    from ai.backend.manager.api.gql.rbac.types.permission import (
        PermissionConnection,
        PermissionFilter,
        PermissionOrderBy,
    )
    from ai.backend.manager.api.gql.user.types.node import UserV2GQL

# ==================== Enums ====================


@strawberry.enum(name="RoleSource", description="Added in 26.3.0. Role definition source")
class RoleSourceGQL(StrEnum):
    SYSTEM = "system"
    CUSTOM = "custom"

    @classmethod
    def from_internal(cls, value: RoleSource) -> RoleSourceGQL:
        return cls(value.value)

    def to_internal(self) -> RoleSource:
        return RoleSource(self.value)


@strawberry.enum(name="RoleStatus", description="Added in 26.3.0. Role status")
class RoleStatusGQL(StrEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    DELETED = "deleted"

    @classmethod
    def from_internal(cls, value: RoleStatus) -> RoleStatusGQL:
        return cls(value.value)

    def to_internal(self) -> RoleStatus:
        return RoleStatus(self.value)


@strawberry.enum(description="Added in 26.3.0. Role ordering field")
class RoleOrderField(StrEnum):
    NAME = "name"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


# ==================== Node Types ====================


@strawberry.type(name="Role", description="Added in 26.3.0. RBAC role")
class RoleGQL(Node):
    id: NodeID[str]
    name: str
    description: str | None
    source: RoleSourceGQL
    status: RoleStatusGQL
    created_at: datetime
    updated_at: datetime
    deleted_at: datetime | None

    @classmethod
    async def resolve_nodes(  # type: ignore[override]
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.role_loader.load_many([
            uuid.UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @classmethod
    def from_dataclass(cls, data: RoleData | RoleDetailData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            description=data.description,
            source=RoleSourceGQL.from_internal(data.source),
            status=RoleStatusGQL.from_internal(data.status),
            created_at=data.created_at,
            updated_at=data.updated_at,
            deleted_at=data.deleted_at,
        )

    @strawberry.field(description="Added in 26.3.0. Permissions associated with this role.")  # type: ignore[misc]
    async def permissions(
        self,
        info: Info[StrawberryGQLContext],
        filter: Annotated[
            PermissionFilter,
            strawberry.lazy("ai.backend.manager.api.gql.rbac.types.permission"),
        ]
        | None = None,
        order_by: list[
            Annotated[
                PermissionOrderBy,
                strawberry.lazy("ai.backend.manager.api.gql.rbac.types.permission"),
            ]
        ]
        | None = None,
        before: str | None = None,
        after: str | None = None,
        first: int | None = None,
        last: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> Annotated[
        PermissionConnection,
        strawberry.lazy("ai.backend.manager.api.gql.rbac.types.permission"),
    ]:
        from ai.backend.manager.api.gql.rbac.fetcher.permission import fetch_permissions
        from ai.backend.manager.api.gql.rbac.types.permission import PermissionFilter

        # Add role_id filter to scope permissions to this role
        role_filter = PermissionFilter(role_id=uuid.UUID(self.id))
        if filter is not None:
            # Merge with user-provided filter
            combined_filter = PermissionFilter(
                role_id=role_filter.role_id,
                scope_type=filter.scope_type,
                entity_type=filter.entity_type,
            )
        else:
            combined_filter = role_filter

        return await fetch_permissions(
            info,
            filter=combined_filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
        )

    @strawberry.field(description="Added in 26.3.0. Users assigned to this role.")  # type: ignore[misc]
    async def users(
        self,
        info: Info[StrawberryGQLContext],
        filter: RoleAssignmentFilter | None = None,
        order_by: list[RoleAssignmentOrderBy] | None = None,
        before: str | None = None,
        after: str | None = None,
        first: int | None = None,
        last: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
    ) -> RoleAssignmentConnection:
        from ai.backend.manager.api.gql.rbac.fetcher.role import fetch_role_assignments

        # Add role_id filter to scope assignments to this role
        role_filter = RoleAssignmentFilter(role_id=uuid.UUID(self.id))
        if filter is not None:
            # Merge with user-provided filter
            combined_filter = RoleAssignmentFilter(
                role_id=role_filter.role_id,
                role=filter.role,
                username=filter.username,
                email=filter.email,
            )
        else:
            combined_filter = role_filter

        return await fetch_role_assignments(
            info,
            filter=combined_filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
        )


@strawberry.type(
    name="RoleAssignment",
    description="Added in 26.3.0. RBAC role assignment (user-role association)",
)
class RoleAssignmentGQL(Node):
    id: NodeID[str]
    user_id: uuid.UUID = strawberry.field(description="The assigned user ID.")
    role_id: uuid.UUID = strawberry.field(description="The assigned role ID.")
    granted_by: uuid.UUID | None = strawberry.field(
        description="The user who granted this assignment."
    )
    granted_at: datetime

    @classmethod
    async def resolve_nodes(  # type: ignore[override]
        cls,
        *,
        info: Info[StrawberryGQLContext],
        node_ids: Iterable[str],
        required: bool = False,
    ) -> Iterable[Self | None]:
        results = await info.context.data_loaders.role_assignment_loader.load_many([
            uuid.UUID(nid) for nid in node_ids
        ])
        return [cls.from_dataclass(data) if data is not None else None for data in results]

    @strawberry.field(description="The assigned role.")  # type: ignore[misc]
    async def role(self, info: Info[StrawberryGQLContext]) -> RoleGQL | None:
        data = await info.context.data_loaders.role_loader.load(self.role_id)
        if data is None:
            return None
        return RoleGQL.from_dataclass(data)

    @strawberry.field(description="The assigned user.")  # type: ignore[misc]
    async def user(
        self, info: Info[StrawberryGQLContext]
    ) -> (
        Annotated[
            UserV2GQL,
            strawberry.lazy("ai.backend.manager.api.gql.user.types.node"),
        ]
        | None
    ):
        from ai.backend.manager.api.gql.user.types.node import UserV2GQL

        data = await info.context.data_loaders.user_loader.load(self.user_id)
        if data is None:
            return None
        return UserV2GQL.from_data(data)

    @classmethod
    def from_dataclass(cls, data: AssignedUserData) -> Self:
        return cls(
            id=ID(str(data.id)),
            user_id=data.user_id,
            role_id=data.role_id,
            granted_by=data.granted_by,
            granted_at=data.granted_at,
        )

    @classmethod
    def from_assignment_data(cls, data: UserRoleAssignmentData) -> Self:
        return cls(
            id=ID(str(data.id)),
            user_id=data.user_id,
            role_id=data.role_id,
            granted_by=data.granted_by,
            granted_at=datetime.now(tz=UTC),
        )

    @classmethod
    def from_revocation_data(cls, data: UserRoleRevocationData) -> Self:
        return cls(
            id=ID(str(data.user_role_id)),
            user_id=data.user_id,
            role_id=data.role_id,
            granted_by=None,
            granted_at=datetime.now(tz=UTC),
        )


# ==================== Filter Types ====================


@strawberry.input(description="Added in 26.3.0. Filter for roles")
class RoleFilter(GQLFilter):
    name: StringFilter | None = None
    source: list[RoleSourceGQL] | None = None
    status: list[RoleStatusGQL] | None = None

    AND: list[RoleFilter] | None = None
    OR: list[RoleFilter] | None = None
    NOT: RoleFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []

        if self.name is not None:
            condition = self.name.build_query_condition(
                contains_factory=RoleConditions.by_name_contains,
                equals_factory=RoleConditions.by_name_equals,
                starts_with_factory=RoleConditions.by_name_starts_with,
                ends_with_factory=RoleConditions.by_name_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.source is not None and len(self.source) > 0:
            conditions.append(RoleConditions.by_sources([s.to_internal() for s in self.source]))

        if self.status is not None and len(self.status) > 0:
            conditions.append(RoleConditions.by_statuses([s.to_internal() for s in self.status]))

        return conditions


# TODO: Add user_id filter (requires AssignedUserConditions.by_user_id)


@strawberry.input(
    name="RoleAssignmentRoleNestedFilter",
    description=(
        "Added in 26.3.0. Nested filter for roles within a role assignment. "
        "Filters assignments that have a role matching all specified conditions."
    ),
)
class RoleAssignmentRoleNestedFilterGQL:
    name: StringFilter | None = None
    source: list[RoleSourceGQL] | None = None
    status: list[RoleStatusGQL] | None = None

    def build_conditions(self) -> list[QueryCondition]:
        raw_conditions: list[QueryCondition] = []
        if self.name:
            condition = self.name.build_query_condition(
                contains_factory=RoleConditions.by_name_contains,
                equals_factory=RoleConditions.by_name_equals,
                starts_with_factory=RoleConditions.by_name_starts_with,
                ends_with_factory=RoleConditions.by_name_ends_with,
            )
            if condition:
                raw_conditions.append(condition)
        if self.source is not None and len(self.source) > 0:
            raw_conditions.append(RoleConditions.by_sources([s.to_internal() for s in self.source]))
        if self.status is not None and len(self.status) > 0:
            raw_conditions.append(
                RoleConditions.by_statuses([s.to_internal() for s in self.status])
            )
        if not raw_conditions:
            return []
        return [AssignedUserConditions.exists_role_combined(raw_conditions)]


@strawberry.input(description="Added in 26.3.0. Filter for role assignments")
class RoleAssignmentFilter(GQLFilter):
    role_id: uuid.UUID | None = None
    role: RoleAssignmentRoleNestedFilterGQL | None = None
    username: StringFilter | None = None
    email: StringFilter | None = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []

        if self.role_id is not None:
            conditions.append(AssignedUserConditions.by_role_id(self.role_id))

        if self.role:
            conditions.extend(self.role.build_conditions())

        if self.username is not None:
            condition = self.username.build_query_condition(
                contains_factory=AssignedUserConditions.by_username_contains,
                equals_factory=AssignedUserConditions.by_username_equals,
                starts_with_factory=AssignedUserConditions.by_username_starts_with,
                ends_with_factory=AssignedUserConditions.by_username_ends_with,
            )
            if condition:
                conditions.append(condition)

        if self.email is not None:
            condition = self.email.build_query_condition(
                contains_factory=AssignedUserConditions.by_email_contains,
                equals_factory=AssignedUserConditions.by_email_equals,
                starts_with_factory=AssignedUserConditions.by_email_starts_with,
                ends_with_factory=AssignedUserConditions.by_email_ends_with,
            )
            if condition:
                conditions.append(condition)

        return conditions


# ==================== OrderBy Types ====================


@strawberry.input(description="Added in 26.3.0. Order by specification for roles")
class RoleOrderBy(GQLOrderBy):
    field: RoleOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case RoleOrderField.NAME:
                return RoleOrders.name(ascending)
            case RoleOrderField.CREATED_AT:
                return RoleOrders.created_at(ascending)
            case RoleOrderField.UPDATED_AT:
                return RoleOrders.updated_at(ascending)


@strawberry.enum(description="Added in 26.3.0. Role assignment ordering field")
class RoleAssignmentOrderField(StrEnum):
    USERNAME = "username"
    EMAIL = "email"
    GRANTED_AT = "granted_at"


@strawberry.input(description="Added in 26.3.0. Order by specification for role assignments")
class RoleAssignmentOrderBy(GQLOrderBy):
    field: RoleAssignmentOrderField
    direction: OrderDirection = OrderDirection.DESC

    @override
    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case RoleAssignmentOrderField.USERNAME:
                return AssignedUserOrders.username(ascending)
            case RoleAssignmentOrderField.EMAIL:
                return AssignedUserOrders.email(ascending)
            case RoleAssignmentOrderField.GRANTED_AT:
                return AssignedUserOrders.granted_at(ascending)


# ==================== Input Types ====================


@strawberry.input(description="Added in 26.3.0. Input for creating a role")
class CreateRoleInput:
    name: str
    description: str | None = None
    source: RoleSourceGQL | None = None

    def to_creator(self) -> Creator[RoleRow]:
        return Creator(
            spec=RoleCreatorSpec(
                name=self.name,
                source=self.source.to_internal() if self.source is not None else RoleSource.CUSTOM,
                status=RoleStatus.ACTIVE,
                description=self.description,
            )
        )


@strawberry.input(description="Added in 26.3.0. Input for updating a role")
class UpdateRoleInput:
    id: uuid.UUID
    name: str | None = None
    description: str | None = None
    status: RoleStatusGQL | None = None

    def to_updater(self) -> Updater[RoleRow]:
        spec = RoleUpdaterSpec(
            name=OptionalState.update(self.name) if self.name is not None else OptionalState.nop(),
            description=(
                TriState.update(self.description)
                if self.description is not None
                else TriState.nop()
            ),
            status=(
                OptionalState.update(self.status.to_internal())
                if self.status is not None
                else OptionalState.nop()
            ),
        )
        return Updater(spec=spec, pk_value=self.id)


@strawberry.input(description="Added in 26.3.0. Input for assigning a role to a user")
class AssignRoleInput:
    user_id: uuid.UUID
    role_id: uuid.UUID

    def to_input(self) -> UserRoleAssignmentInput:
        return UserRoleAssignmentInput(
            user_id=self.user_id,
            role_id=self.role_id,
        )


@strawberry.input(description="Added in 26.3.0. Input for revoking a role from a user")
class RevokeRoleInput:
    user_id: uuid.UUID
    role_id: uuid.UUID

    def to_input(self) -> UserRoleRevocationInput:
        return UserRoleRevocationInput(
            user_id=self.user_id,
            role_id=self.role_id,
        )


@strawberry.input(
    name="BulkAssignRoleInput",
    description="Added in 26.3.0. Input for bulk assigning a role to multiple users",
)
class BulkAssignRoleInputGQL:
    role_id: uuid.UUID
    user_ids: list[uuid.UUID]

    def to_bulk_creator(self) -> BulkCreator[UserRoleRow]:
        specs = [UserRoleCreatorSpec(user_id=uid, role_id=self.role_id) for uid in self.user_ids]
        return BulkCreator(specs=specs)


@strawberry.input(
    name="BulkRevokeRoleInput",
    description="Added in 26.3.0. Input for bulk revoking a role from multiple users",
)
class BulkRevokeRoleInputGQL:
    role_id: uuid.UUID
    user_ids: list[uuid.UUID]

    def to_input(self) -> BulkUserRoleRevocationInput:
        return BulkUserRoleRevocationInput(
            role_id=self.role_id,
            user_ids=self.user_ids,
        )


@strawberry.input(description="Added in 26.3.0. Input for soft-deleting a role")
class DeleteRoleInput:
    id: uuid.UUID


@strawberry.input(description="Added in 26.3.0. Input for purging a role")
class PurgeRoleInput:
    id: uuid.UUID


# ==================== Payload Types ====================


@strawberry.type(description="Added in 26.3.0. Payload for delete role mutation")
class DeleteRolePayload:
    id: ID


@strawberry.type(description="Added in 26.3.0. Payload for purge role mutation")
class PurgeRolePayload:
    id: ID


@strawberry.type(
    name="BulkAssignRoleError",
    description="Added in 26.3.0. Error information for a failed user in bulk role assignment.",
)
class BulkAssignRoleErrorGQL:
    user_id: uuid.UUID = strawberry.field(description="UUID of the user that failed.")
    message: str = strawberry.field(description="Error message describing the failure.")


@strawberry.type(
    name="BulkAssignRolePayload",
    description="Added in 26.3.0. Payload for bulk role assignment mutation.",
)
class BulkAssignRolePayloadGQL:
    assigned: list[RoleAssignmentGQL] = strawberry.field(
        description="List of successfully created role assignments."
    )
    failed: list[BulkAssignRoleErrorGQL] = strawberry.field(
        description="List of errors for users that failed to be assigned."
    )


@strawberry.type(
    name="BulkRevokeRoleError",
    description="Added in 26.3.0. Error information for a failed user in bulk role revocation.",
)
class BulkRevokeRoleErrorGQL:
    user_id: uuid.UUID = strawberry.field(description="UUID of the user that failed.")
    message: str = strawberry.field(description="Error message describing the failure.")


@strawberry.type(
    name="BulkRevokeRolePayload",
    description="Added in 26.3.0. Payload for bulk role revocation mutation.",
)
class BulkRevokeRolePayloadGQL:
    revoked: list[RoleAssignmentGQL] = strawberry.field(
        description="List of successfully revoked role assignments."
    )
    failed: list[BulkRevokeRoleErrorGQL] = strawberry.field(
        description="List of errors for users that failed to be revoked."
    )


# ==================== Connection Types ====================


RoleEdge = Edge[RoleGQL]


@strawberry.type(description="Added in 26.3.0. Role connection")
class RoleConnection(Connection[RoleGQL]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


RoleAssignmentEdge = Edge[RoleAssignmentGQL]


@strawberry.type(description="Added in 26.3.0. Role assignment connection")
class RoleAssignmentConnection(Connection[RoleAssignmentGQL]):
    count: int

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Sequence

from ai.backend.manager.data.deployment.types import (
    DeploymentLifecycleStatus,
    DeploymentStatusTransitions,
)
from ai.backend.manager.defs import LockID
from ai.backend.manager.sokovan.deployment.types import (
    DeploymentExecutionResult,
    DeploymentWithHistory,
)


class DeploymentHandler:
    """Base class for deployment operation handlers.

    Each handler targets deployments matching specific lifecycle statuses
    and optionally a sub-step.  The coordinator queries the DB using these
    filters and passes only the matching deployments to ``execute()``.

    ``target_statuses()`` fully describes what each handler processes:
    the coordinator derives both the DB query filter and the registry key
    from the returned ``DeploymentLifecycleStatus`` list.
    """

    @classmethod
    @abstractmethod
    def name(cls) -> str:
        """Get the name of the handler."""
        raise NotImplementedError("Subclasses must implement name()")

    @property
    @abstractmethod
    def lock_id(self) -> LockID | None:
        """Get the lock ID for this handler.

        Returns:
            LockID to acquire before execution, or None if no lock needed
        """
        raise NotImplementedError("Subclasses must implement lock_id")

    @classmethod
    @abstractmethod
    def target_statuses(cls) -> list[DeploymentLifecycleStatus]:
        """Get the target deployment statuses for this handler.

        Each entry pairs an ``EndpointLifecycle`` with an optional
        ``DeploymentSubStatus`` (sub-step).  The coordinator uses these
        to filter deployments from the DB and to build registry keys.

        Returns:
            List of lifecycle + sub-step pairs that this handler targets
        """
        raise NotImplementedError("Subclasses must implement target_statuses()")

    @classmethod
    @abstractmethod
    def status_transitions(cls) -> DeploymentStatusTransitions:
        """Define state transitions for different handler outcomes (BEP-1030).

        Returns:
            DeploymentStatusTransitions defining what lifecycle to transition to for
            success and failure outcomes.

        Note:
            - None value: Don't change the deployment lifecycle
        """
        raise NotImplementedError("Subclasses must implement status_transitions()")

    @abstractmethod
    async def execute(
        self, deployments: Sequence[DeploymentWithHistory]
    ) -> DeploymentExecutionResult:
        """Execute the scheduling operation.

        Args:
            deployments: Deployments bundled with scheduling history context

        Returns:
            Result of the scheduling operation
        """
        raise NotImplementedError("Subclasses must implement execute()")

    @abstractmethod
    async def post_process(self, result: DeploymentExecutionResult) -> None:
        """Per-handler post-processing after execute().

        Typical use: reschedule the next lifecycle cycle, trigger dependent lifecycles.

        Args:
            result: The result from this handler's execute()
        """
        raise NotImplementedError("Subclasses must implement post_process()")



"""Blue-green deployment strategy evaluation for a single deployment cycle (BEP-1049).

Provisions a full set of new-revision routes, validates them, then atomically
switches traffic from the old revision to the new one.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import override

from ai.backend.manager.data.deployment.types import (
    DeploymentInfo,
    RouteInfo,
)
from ai.backend.manager.models.deployment_policy import BlueGreenSpec

from .types import AbstractDeploymentStrategy, StrategyCycleResult


class BlueGreenStrategy(AbstractDeploymentStrategy):
    """Blue-green deployment strategy FSM."""

    def __init__(self, spec: BlueGreenSpec) -> None:
        super().__init__(spec)
        self._spec = spec

    @override
    def evaluate_cycle(
        self,
        deployment: DeploymentInfo,
        routes: Sequence[RouteInfo],
    ) -> StrategyCycleResult:
        """Evaluate one cycle of blue-green deployment for a single deployment."""
        raise NotImplementedError("Blue-green deployment strategy is not yet implemented")



from ai.backend.common.types import AccessKey
from ai.backend.manager.models.session import KernelLoadingStrategy, SessionRow
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


class StreamDBSource:
    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def get_streaming_session(
        self,
        session_name: str,
        access_key: AccessKey,
    ) -> SessionRow:
        async with self._db.begin_readonly_session() as db_sess:
            return await SessionRow.get_session(
                db_sess,
                session_name,
                access_key,
                kernel_loading_strategy=KernelLoadingStrategy.MAIN_KERNEL_ONLY,
            )



import uuid
from dataclasses import dataclass, field
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.model_serving.types import ServiceSearchItem
from ai.backend.manager.repositories.base.types import QueryCondition
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class SearchServicesAction(ModelServiceAction):
    session_owner_id: uuid.UUID
    conditions: list[QueryCondition] = field(default_factory=list)
    offset: int = 0
    limit: int = 20

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchServicesActionResult(BaseActionResult):
    items: list[ServiceSearchItem]
    total_count: int
    offset: int
    limit: int

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

from ai.backend.common.dependencies import NonMonitorableDependencyProvider, ResourceT, SetupInputT


class DomainDependency(NonMonitorableDependencyProvider[SetupInputT, ResourceT]):
    """Base class for domain dependencies.

    Domain dependencies are higher-level objects (repositories, services, lock factories)
    that depend on infrastructure (db, valkey, etcd) and component (storage_manager) resources.
    They use composite input dataclasses rather than a single config object.
    """

    pass



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.clients.prometheus.client import PrometheusClient
from ai.backend.common.clients.valkey_client.valkey_artifact.client import (
    ValkeyArtifactDownloadTrackingClient,
)
from ai.backend.common.clients.valkey_client.valkey_container_log.client import (
    ValkeyContainerLogClient,
)
from ai.backend.common.clients.valkey_client.valkey_live.client import ValkeyLiveClient
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.clients.valkey_client.valkey_stream.client import ValkeyStreamClient
from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventDispatcher, EventObserver, EventProducer
from ai.backend.common.events.fetcher import EventFetcher
from ai.backend.common.events.hub.hub import EventHub
from ai.backend.common.message_queue.abc.queue import AbstractMessageQueue
from ai.backend.common.plugin.event import EventDispatcherPluginContext
from ai.backend.common.plugin.hook import HookPluginContext
from ai.backend.common.plugin.monitor import ErrorPluginContext, StatsPluginContext
from ai.backend.manager.actions.monitors.audit_log import AuditLogMonitor
from ai.backend.manager.actions.monitors.prometheus import PrometheusMonitor
from ai.backend.manager.actions.monitors.reporter import ReporterMonitor
from ai.backend.manager.actions.validators import ActionValidators
from ai.backend.manager.actions.validators.rbac import RBACValidators
from ai.backend.manager.actions.validators.rbac.scope import ScopeActionRBACValidator
from ai.backend.manager.actions.validators.rbac.single_entity import (
    SingleEntityActionRBACValidator,
)
from ai.backend.manager.agent_cache import AgentRPCCache
from ai.backend.manager.clients.agent.pool import AgentClientPool
from ai.backend.manager.clients.appproxy.client import AppProxyClientPool
from ai.backend.manager.clients.storage_proxy.session_manager import StorageSessionManager
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.event_dispatcher.dispatch import DispatcherArgs, Dispatchers
from ai.backend.manager.event_dispatcher.handlers.stream_cleanup import StreamCleanupEventHandler
from ai.backend.manager.idle import IdleCheckerHost
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.notification import NotificationCenter
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.reporters.base import AbstractReporter
from ai.backend.manager.reporters.hub import ReporterHub, ReporterHubArgs
from ai.backend.manager.reporters.smtp import SMTPReporter, SMTPSenderArgs
from ai.backend.manager.repositories.repositories import Repositories
from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.service.container_registry.harbor import (
    AbstractPerProjectContainerRegistryQuotaService,
)
from ai.backend.manager.services.processors import Processors, ServiceArgs
from ai.backend.manager.sokovan.deployment import DeploymentController
from ai.backend.manager.sokovan.deployment.coordinator import DeploymentCoordinator
from ai.backend.manager.sokovan.deployment.revision_generator.registry import (
    RevisionGeneratorRegistry,
)
from ai.backend.manager.sokovan.deployment.route.coordinator import RouteCoordinator
from ai.backend.manager.sokovan.scheduler.coordinator import ScheduleCoordinator
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController
from ai.backend.manager.types import DistributedLockFactory, SMTPTriggerPolicy

from .agent_lost_checker import AgentLostCheckerDependency, AgentLostCheckerInput
from .bgtask_registry import BgtaskRegistryDependency, BgtaskRegistryInput
from .event_dispatcher import EventDispatcherDependency, EventDispatcherInput
from .log_cleanup_timer import LogCleanupTimerDependency, LogCleanupTimerInput
from .manager_status_watcher import ManagerStatusWatcherDependency, ManagerStatusWatcherInput
from .processors import ProcessorsDependency, ProcessorsProviderInput
from .stats_reporter import StatsReporterDependency, StatsReporterInput


@dataclass
class ProcessingInput:
    """Input required for processing layer setup.

    Contains all dependencies needed for event dispatcher,
    processors, and background task registry.
    """

    # EventDispatcher creation
    message_queue: AbstractMessageQueue
    log_events: bool
    event_observer: EventObserver | None

    # Dispatchers registration (DispatcherArgs)
    valkey_container_log: ValkeyContainerLogClient
    valkey_stat: ValkeyStatClient
    valkey_stream: ValkeyStreamClient
    schedule_coordinator: ScheduleCoordinator
    scheduling_controller: SchedulingController
    deployment_coordinator: DeploymentCoordinator
    route_coordinator: RouteCoordinator
    scheduler_repository: SchedulerRepository
    event_hub: EventHub
    agent_registry: AgentRegistry
    db: ExtendedAsyncSAEngine
    idle_checker_host: IdleCheckerHost
    event_dispatcher_plugin_ctx: EventDispatcherPluginContext
    repositories: Repositories
    storage_manager: StorageSessionManager
    config_provider: ManagerConfigProvider
    event_producer: EventProducer

    # Processors creation (ServiceArgs additional)
    etcd: AsyncEtcd
    valkey_live: ValkeyLiveClient
    valkey_artifact_client: ValkeyArtifactDownloadTrackingClient
    event_fetcher: EventFetcher
    background_task_manager: BackgroundTaskManager
    error_monitor: ErrorPluginContext
    hook_plugin_ctx: HookPluginContext
    deployment_controller: DeploymentController
    revision_generator_registry: RevisionGeneratorRegistry
    agent_cache: AgentRPCCache
    notification_center: NotificationCenter
    appproxy_client_pool: AppProxyClientPool
    prometheus_client: PrometheusClient

    # BgtaskRegistry creation (additional)
    agent_client_pool: AgentClientPool

    # Log cleanup timer
    distributed_lock_factory: DistributedLockFactory

    # Lifecycle background tasks
    stats_monitor: StatsPluginContext
    pidx: int

    # Registry quota service (optional, defaults to None)
    registry_quota_service: AbstractPerProjectContainerRegistryQuotaService | None = None


@dataclass
class ProcessingResources:
    """Container for processing layer resources."""

    event_dispatcher: EventDispatcher
    processors: Processors
    stream_cleanup_handler: StreamCleanupEventHandler


def _make_registered_reporters(
    config_provider: ManagerConfigProvider,
) -> dict[str, AbstractReporter]:
    reporters: dict[str, AbstractReporter] = {}
    smtp_configs = config_provider.config.reporter.smtp
    for smtp_conf in smtp_configs:
        smtp_args = SMTPSenderArgs(
            host=smtp_conf.host,
            port=smtp_conf.port,
            username=smtp_conf.username,
            password=smtp_conf.password,
            sender=smtp_conf.sender,
            recipients=smtp_conf.recipients,
            use_tls=smtp_conf.use_tls,
            max_workers=smtp_conf.max_workers,
            template=smtp_conf.template,
        )
        trigger_policy = SMTPTriggerPolicy[smtp_conf.trigger_policy]
        reporters[smtp_conf.name] = SMTPReporter(smtp_args, trigger_policy)
    return reporters


def _make_action_reporters(
    config_provider: ManagerConfigProvider,
    reporters: dict[str, AbstractReporter],
) -> dict[str, list[AbstractReporter]]:
    action_monitors: dict[str, list[AbstractReporter]] = {}
    action_monitor_configs = config_provider.config.reporter.action_monitors
    for action_monitor_conf in action_monitor_configs:
        reporter_name: str = action_monitor_conf.reporter
        reporter = reporters.get(reporter_name)
        if reporter is None:
            continue
        for action_type in action_monitor_conf.subscribed_actions:
            action_monitors.setdefault(action_type, []).append(reporter)
    return action_monitors


class ProcessingComposer(DependencyComposer[ProcessingInput, ProcessingResources]):
    """Composes event dispatcher, processors, and background task registry.

    Orchestrates the processing layer (Layer 6-7) initialization:
    1. EventDispatcher: Created first (not yet started)
    2. Processors: Created with event_dispatcher reference
    3. Dispatchers: Registered on event_dispatcher, then started
    4. BgtaskRegistry: Created and set on background_task_manager
    """

    @property
    def stage_name(self) -> str:
        return "processing"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: ProcessingInput,
    ) -> AsyncIterator[ProcessingResources]:
        # Step 1: Create EventDispatcher (not started yet)
        event_dispatcher = await stack.enter_dependency(
            EventDispatcherDependency(),
            EventDispatcherInput(
                message_queue=setup_input.message_queue,
                log_events=setup_input.log_events,
                event_observer=setup_input.event_observer,
            ),
        )

        # Step 2: Build action monitors and create Processors
        registered_reporters = _make_registered_reporters(setup_input.config_provider)
        action_reporters = _make_action_reporters(setup_input.config_provider, registered_reporters)
        reporter_hub = ReporterHub(ReporterHubArgs(reporters=action_reporters))
        reporter_monitor = ReporterMonitor(reporter_hub)
        prometheus_monitor = PrometheusMonitor()
        audit_log_monitor = AuditLogMonitor(setup_input.repositories.audit_log.repository)

        service_args = ServiceArgs(
            db=setup_input.db,
            repositories=setup_input.repositories,
            etcd=setup_input.etcd,
            config_provider=setup_input.config_provider,
            storage_manager=setup_input.storage_manager,
            valkey_stat_client=setup_input.valkey_stat,
            valkey_live=setup_input.valkey_live,
            valkey_artifact_client=setup_input.valkey_artifact_client,
            event_fetcher=setup_input.event_fetcher,
            background_task_manager=setup_input.background_task_manager,
            event_hub=setup_input.event_hub,
            agent_registry=setup_input.agent_registry,
            error_monitor=setup_input.error_monitor,
            idle_checker_host=setup_input.idle_checker_host,
            event_dispatcher=event_dispatcher,
            hook_plugin_ctx=setup_input.hook_plugin_ctx,
            scheduling_controller=setup_input.scheduling_controller,
            deployment_controller=setup_input.deployment_controller,
            revision_generator_registry=setup_input.revision_generator_registry,
            event_producer=setup_input.event_producer,
            agent_cache=setup_input.agent_cache,
            notification_center=setup_input.notification_center,
            appproxy_client_pool=setup_input.appproxy_client_pool,
            prometheus_client=setup_input.prometheus_client,
            registry_quota_service=setup_input.registry_quota_service,
        )

        permission_controller_repository = setup_input.repositories.permission_controller.repository
        rbac_validators = RBACValidators(
            scope=ScopeActionRBACValidator(permission_controller_repository),
            single_entity=SingleEntityActionRBACValidator(permission_controller_repository),
        )

        processors = await stack.enter_dependency(
            ProcessorsDependency(),
            ProcessorsProviderInput(
                service_args=service_args,
                action_monitors=[reporter_monitor, prometheus_monitor, audit_log_monitor],
                event_hub=setup_input.event_hub,
                event_fetcher=setup_input.event_fetcher,
                validators=ActionValidators(rbac=rbac_validators),
            ),
        )

        # Step 3: Register Dispatchers and start EventDispatcher
        dispatchers = Dispatchers(
            DispatcherArgs(
                valkey_container_log=setup_input.valkey_container_log,
                valkey_stat=setup_input.valkey_stat,
                valkey_stream=setup_input.valkey_stream,
                schedule_coordinator=setup_input.schedule_coordinator,
                scheduling_controller=setup_input.scheduling_controller,
                deployment_coordinator=setup_input.deployment_coordinator,
                route_coordinator=setup_input.route_coordinator,
                scheduler_repository=setup_input.scheduler_repository,
                event_hub=setup_input.event_hub,
                agent_registry=setup_input.agent_registry,
                db=setup_input.db,
                etcd=setup_input.etcd,
                idle_checker_host=setup_input.idle_checker_host,
                event_dispatcher_plugin_ctx=setup_input.event_dispatcher_plugin_ctx,
                repositories=setup_input.repositories,
                processors_factory=lambda: processors,
                storage_manager=setup_input.storage_manager,
                config_provider=setup_input.config_provider,
                event_producer=setup_input.event_producer,
            )
        )
        dispatchers.dispatch(event_dispatcher)
        await event_dispatcher.start()

        # Step 3.5: Create and start log cleanup timer
        await stack.enter_dependency(
            LogCleanupTimerDependency(),
            LogCleanupTimerInput(
                distributed_lock_factory=setup_input.distributed_lock_factory,
                event_producer=setup_input.event_producer,
            ),
        )

        # Step 4: Create BgtaskRegistry
        await stack.enter_dependency(
            BgtaskRegistryDependency(),
            BgtaskRegistryInput(
                processors=processors,
                background_task_manager=setup_input.background_task_manager,
                repositories=setup_input.repositories,
                agent_client_pool=setup_input.agent_client_pool,
                agent_registry=setup_input.agent_registry,
                event_hub=setup_input.event_hub,
                event_fetcher=setup_input.event_fetcher,
            ),
        )

        # Step 5: Start lifecycle background tasks
        await stack.enter_dependency(
            AgentLostCheckerDependency(),
            AgentLostCheckerInput(
                config_provider=setup_input.config_provider,
                valkey_live=setup_input.valkey_live,
                event_producer=setup_input.event_producer,
            ),
        )
        await stack.enter_dependency(
            StatsReporterDependency(),
            StatsReporterInput(
                stats_monitor=setup_input.stats_monitor,
                registry=setup_input.agent_registry,
                db=setup_input.db,
            ),
        )
        await stack.enter_dependency(
            ManagerStatusWatcherDependency(),
            ManagerStatusWatcherInput(
                config_provider=setup_input.config_provider,
                pidx=setup_input.pidx,
                valkey_stat=setup_input.valkey_stat,
                db=setup_input.db,
            ),
        )

        yield ProcessingResources(
            event_dispatcher=event_dispatcher,
            processors=processors,
            stream_cleanup_handler=dispatchers.stream_cleanup_handler,
        )



from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from ai.backend.common.bgtask.bgtask import BackgroundTaskManager
from ai.backend.common.clients.prometheus.client import PrometheusClient
from ai.backend.common.dependencies import DependencyComposer, DependencyStack
from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.health_checker import HealthProbe
from ai.backend.common.jwt.validator import JWTValidator
from ai.backend.common.metrics.metric import CommonMetricRegistry
from ai.backend.common.service_discovery.service_discovery import (
    ServiceDiscovery,
    ServiceDiscoveryLoop,
)
from ai.backend.common.types import ValkeyProfileTarget
from ai.backend.manager.api.gql.adapter import BaseGQLAdapter
from ai.backend.manager.api.rest.types import CORSOptions
from ai.backend.manager.config.unified import ManagerUnifiedConfig
from ai.backend.manager.dependencies.infrastructure.redis import ValkeyClients
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

from .background_task_manager import BackgroundTaskManagerDependency, BackgroundTaskManagerInput
from .cors_options import CORSOptionsDependency
from .gql_adapter import GQLAdapterDependency
from .health_probe import HealthProbeDependency, HealthProbeInput
from .jwt_validator import JWTValidatorDependency
from .metrics import MetricsDependency
from .prometheus_client import PrometheusClientDependency
from .service_discovery import ServiceDiscoveryDependency, ServiceDiscoveryInput


@dataclass
class SystemInput:
    """Input required for system services setup.

    Contains configuration and infrastructure resources from earlier stages.
    """

    config: ManagerUnifiedConfig
    etcd: AsyncEtcd
    valkey: ValkeyClients
    db: ExtendedAsyncSAEngine
    event_producer: EventProducer
    valkey_profile_target: ValkeyProfileTarget


@dataclass
class SystemResources:
    """Container for all system service resources."""

    cors_options: CORSOptions
    metrics: CommonMetricRegistry
    gql_adapter: BaseGQLAdapter
    jwt_validator: JWTValidator
    prometheus_client: PrometheusClient
    service_discovery: ServiceDiscovery
    sd_loop: ServiceDiscoveryLoop
    background_task_manager: BackgroundTaskManager
    health_probe: HealthProbe


class SystemComposer(DependencyComposer[SystemInput, SystemResources]):
    """Composes all system service dependencies.

    Orchestrates initialization across 4 layers:
    - Layer 0: CORS options, metrics, GQL adapter (no dependencies)
    - Layer 1: JWT validator, Prometheus client, service discovery (config-dependent)
    - Layer 3: Background task manager (event_producer-dependent)
    - Layer 4: Health probe (depends on all infrastructure)
    """

    @property
    def stage_name(self) -> str:
        return "system"

    @asynccontextmanager
    async def compose(
        self,
        stack: DependencyStack,
        setup_input: SystemInput,
    ) -> AsyncIterator[SystemResources]:
        # Layer 0: No external dependencies
        cors_options = await stack.enter_dependency(CORSOptionsDependency(), None)
        metrics = await stack.enter_dependency(MetricsDependency(), None)
        gql_adapter = await stack.enter_dependency(GQLAdapterDependency(), None)

        # Layer 1: Config-dependent services
        jwt_validator = await stack.enter_dependency(JWTValidatorDependency(), setup_input.config)
        prometheus_client = await stack.enter_dependency(
            PrometheusClientDependency(), setup_input.config
        )
        sd_resources = await stack.enter_dependency(
            ServiceDiscoveryDependency(),
            ServiceDiscoveryInput(
                config=setup_input.config,
                etcd=setup_input.etcd,
                valkey_profile_target=setup_input.valkey_profile_target,
            ),
        )

        # Layer 3: Event-producer-dependent services
        background_task_manager = await stack.enter_dependency(
            BackgroundTaskManagerDependency(),
            BackgroundTaskManagerInput(
                event_producer=setup_input.event_producer,
                valkey_bgtask=setup_input.valkey.bgtask,
                server_id=setup_input.config.manager.id,
                bgtask_observer=metrics.bgtask,
            ),
        )

        # Layer 4: Infrastructure-dependent services
        health_probe = await stack.enter_dependency(
            HealthProbeDependency(),
            HealthProbeInput(
                db=setup_input.db,
                etcd=setup_input.etcd,
                valkey=setup_input.valkey,
            ),
        )

        yield SystemResources(
            cors_options=cors_options,
            metrics=metrics,
            gql_adapter=gql_adapter,
            jwt_validator=jwt_validator,
            prometheus_client=prometheus_client,
            service_discovery=sd_resources.service_discovery,
            sd_loop=sd_resources.sd_loop,
            background_task_manager=background_task_manager,
            health_probe=health_probe,
        )



"""GraphQL resolvers for RBAC permission management."""

from __future__ import annotations

import strawberry
from strawberry import ID, Info

from ai.backend.common.data.permission.scope_entity_combinations import (
    VALID_SCOPE_ENTITY_COMBINATIONS,
)
from ai.backend.manager.api.gql.rbac.fetcher.permission import fetch_permissions
from ai.backend.manager.api.gql.rbac.types import (
    CreatePermissionInput,
    DeletePermissionInput,
    DeletePermissionPayload,
    PermissionConnection,
    PermissionFilter,
    PermissionGQL,
    PermissionOrderBy,
    RBACElementTypeGQL,
    ScopeEntityCombinationGQL,
    UpdatePermissionInput,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.utils import check_admin_only
from ai.backend.manager.models.rbac_models.permission.permission import PermissionRow
from ai.backend.manager.repositories.base.purger import Purger
from ai.backend.manager.services.permission_contoller.actions.permission import (
    CreatePermissionAction,
    DeletePermissionAction,
)
from ai.backend.manager.services.permission_contoller.actions.update_permission import (
    UpdatePermissionAction,
)

# ==================== Query Resolvers ====================


@strawberry.field(
    description="Added in 26.3.0. List scoped permissions with filtering and pagination (admin only)."
)  # type: ignore[misc]
async def admin_permissions(
    info: Info[StrawberryGQLContext],
    filter: PermissionFilter | None = None,
    order_by: list[PermissionOrderBy] | None = None,
    before: str | None = None,
    after: str | None = None,
    first: int | None = None,
    last: int | None = None,
    limit: int | None = None,
    offset: int | None = None,
) -> PermissionConnection:
    check_admin_only()
    return await fetch_permissions(
        info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(description="Added in 26.3.0. List valid RBAC scope-entity type combinations.")  # type: ignore[misc]
async def rbac_scope_entity_combinations(
    info: Info[StrawberryGQLContext],
) -> list[ScopeEntityCombinationGQL]:
    return [
        ScopeEntityCombinationGQL(
            scope_type=RBACElementTypeGQL.from_element(scope),
            valid_entity_types=sorted(
                [RBACElementTypeGQL.from_element(entity) for entity in entities],
                key=lambda e: e.value,
            ),
        )
        for scope, entities in VALID_SCOPE_ENTITY_COMBINATIONS.items()
    ]


# ==================== Mutation Resolvers ====================


@strawberry.mutation(description="Added in 26.3.0. Create a scoped permission (admin only).")  # type: ignore[misc]
async def admin_create_permission(
    info: Info[StrawberryGQLContext],
    input: CreatePermissionInput,
) -> PermissionGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.create_permission.wait_for_complete(
            CreatePermissionAction(creator=input.to_creator())
        )
    )
    return PermissionGQL.from_dataclass(action_result.data)


@strawberry.mutation(description="Added in 26.3.0. Update a scoped permission (admin only).")  # type: ignore[misc]
async def admin_update_permission(
    info: Info[StrawberryGQLContext],
    input: UpdatePermissionInput,
) -> PermissionGQL:
    check_admin_only()
    action_result = (
        await info.context.processors.permission_controller.update_permission.wait_for_complete(
            UpdatePermissionAction(updater=input.to_updater())
        )
    )
    return PermissionGQL.from_dataclass(action_result.data)


@strawberry.mutation(description="Added in 26.3.0. Delete a scoped permission (admin only).")  # type: ignore[misc]
async def admin_delete_permission(
    info: Info[StrawberryGQLContext],
    input: DeletePermissionInput,
) -> DeletePermissionPayload:
    check_admin_only()
    purger = Purger(row_class=PermissionRow, pk_value=input.id)
    await info.context.processors.permission_controller.delete_permission.wait_for_complete(
        DeletePermissionAction(purger=purger)
    )
    return DeletePermissionPayload(id=ID(str(input.id)))



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import Any
from uuid import UUID

from ai.backend.common.types import ServiceCatalogStatus


@dataclass
class ServiceCatalogEndpointData:
    id: UUID
    service_id: UUID
    role: str
    scope: str
    address: str
    port: int
    protocol: str
    metadata: dict[str, Any] | None = field(default=None)


@dataclass
class ServiceCatalogData:
    id: UUID
    service_group: str
    instance_id: str
    display_name: str
    version: str
    labels: dict[str, Any]
    status: ServiceCatalogStatus
    startup_time: datetime
    registered_at: datetime
    last_heartbeat: datetime
    config_hash: str
    endpoints: list[ServiceCatalogEndpointData] = field(default_factory=list)



from __future__ import annotations

import collections
from collections.abc import Mapping
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast

from ai.backend.common.json import load_json
from ai.backend.common.types import AcceleratorMetadata
from ai.backend.manager.errors.api import InvalidAPIParameters

from .actions.delete_config import DeleteConfigAction, DeleteConfigActionResult
from .actions.get_config import GetConfigAction, GetConfigActionResult
from .actions.get_resource_metadata import (
    GetResourceMetadataAction,
    GetResourceMetadataActionResult,
)
from .actions.get_resource_slots import GetResourceSlotsAction, GetResourceSlotsActionResult
from .actions.get_vfolder_types import GetVfolderTypesAction, GetVfolderTypesActionResult
from .actions.set_config import SetConfigAction, SetConfigActionResult

if TYPE_CHECKING:
    from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
    from ai.backend.common.etcd import AsyncEtcd
    from ai.backend.manager.config.provider import ManagerConfigProvider
    from ai.backend.manager.repositories.etcd_config import EtcdConfigRepository

__all__ = ("EtcdConfigService",)

KNOWN_SLOT_METADATA: dict[str, AcceleratorMetadata] = {
    "cpu": {
        "slot_name": "cpu",
        "description": "CPU",
        "human_readable_name": "CPU",
        "display_unit": "Core",
        "number_format": {"binary": False, "round_length": 0},
        "display_icon": "cpu",
    },
    "mem": {
        "slot_name": "ram",
        "description": "Memory",
        "human_readable_name": "RAM",
        "display_unit": "GiB",
        "number_format": {"binary": True, "round_length": 0},
        "display_icon": "cpu",
    },
    "cuda.device": {
        "slot_name": "cuda.device",
        "human_readable_name": "GPU",
        "description": "CUDA-capable GPU",
        "display_unit": "GPU",
        "number_format": {"binary": False, "round_length": 0},
        "display_icon": "gpu1",
    },
    "cuda.shares": {
        "slot_name": "cuda.shares",
        "human_readable_name": "fGPU",
        "description": "CUDA-capable GPU (fractional)",
        "display_unit": "fGPU",
        "number_format": {"binary": False, "round_length": 2},
        "display_icon": "gpu1",
    },
    "rocm.device": {
        "slot_name": "rocm.device",
        "human_readable_name": "GPU",
        "description": "ROCm-capable GPU",
        "display_unit": "GPU",
        "number_format": {"binary": False, "round_length": 0},
        "display_icon": "gpu2",
    },
    "tpu.device": {
        "slot_name": "tpu.device",
        "human_readable_name": "TPU",
        "description": "TPU device",
        "display_unit": "GPU",
        "number_format": {"binary": False, "round_length": 0},
        "display_icon": "tpu",
    },
}


@dataclass
class EtcdConfigService:
    """Service for etcd configuration operations."""

    _repository: EtcdConfigRepository
    _config_provider: ManagerConfigProvider
    _etcd: AsyncEtcd
    _valkey_stat: ValkeyStatClient

    def __init__(
        self,
        *,
        repository: EtcdConfigRepository,
        config_provider: ManagerConfigProvider,
        etcd: AsyncEtcd,
        valkey_stat: ValkeyStatClient,
    ) -> None:
        self._repository = repository
        self._config_provider = config_provider
        self._etcd = etcd
        self._valkey_stat = valkey_stat

    async def get_resource_slots(
        self, action: GetResourceSlotsAction
    ) -> GetResourceSlotsActionResult:
        """Get system-wide known resource slots."""
        known_slots = await self._config_provider.legacy_etcd_config_loader.get_resource_slots()
        return GetResourceSlotsActionResult(
            slots={str(k): v for k, v in known_slots.items()},
        )

    async def get_resource_metadata(
        self, action: GetResourceMetadataAction
    ) -> GetResourceMetadataActionResult:
        """Get resource metadata with optional scaling group filter."""
        known_slots = await self._config_provider.legacy_etcd_config_loader.get_resource_slots()

        # Collect plugin-reported accelerator metadata
        computer_metadata = await self._valkey_stat.get_computer_metadata()
        reported_accelerator_metadata: dict[str, AcceleratorMetadata] = {
            slot_name: cast(AcceleratorMetadata, load_json(metadata_json))
            for slot_name, metadata_json in computer_metadata.items()
        }

        # Merge reported metadata and preconfigured metadata (for legacy plugins)
        accelerator_metadata: dict[str, AcceleratorMetadata] = {}
        for slot_name, metadata in collections.ChainMap(
            reported_accelerator_metadata,
            KNOWN_SLOT_METADATA,
        ).items():
            if slot_name in known_slots:
                accelerator_metadata[slot_name] = metadata

        # Optionally filter by the slots reported by the given resource group's agents
        if action.sgroup is not None:
            available_slot_keys = await self._repository.get_available_agent_slots(action.sgroup)
            accelerator_metadata = {
                str(k): v
                for k, v in accelerator_metadata.items()
                if k in {"cpu", "mem", *available_slot_keys}
            }

        return GetResourceMetadataActionResult(metadata=accelerator_metadata)

    async def get_vfolder_types(self, action: GetVfolderTypesAction) -> GetVfolderTypesActionResult:
        """Get available vfolder types."""
        vfolder_types = await self._config_provider.legacy_etcd_config_loader.get_vfolder_types()
        return GetVfolderTypesActionResult(types=list(vfolder_types))

    async def get_config(self, action: GetConfigAction) -> GetConfigActionResult:
        """Get raw etcd key-value."""
        if action.prefix:
            tree_value = dict(await self._etcd.get_prefix_dict(action.key))
            return GetConfigActionResult(result=tree_value)
        scalar_value = await self._etcd.get(action.key)
        return GetConfigActionResult(result=scalar_value)

    async def set_config(self, action: SetConfigAction) -> SetConfigActionResult:
        """Set raw etcd key-value."""
        if isinstance(action.value, Mapping):
            updates: dict[str, Any] = {}

            def flatten(prefix: str, o: Mapping[str, Any]) -> None:
                for k, v in o.items():
                    inner_prefix = prefix if k == "" else f"{prefix}/{k}"
                    if isinstance(v, Mapping):
                        flatten(inner_prefix, v)
                    else:
                        updates[inner_prefix] = v

            flatten(action.key, action.value)
            if len(updates) > 16:
                raise InvalidAPIParameters(
                    "Too large update! Split into smaller key-value pair sets."
                )
            await self._etcd.put_dict(updates)
        else:
            await self._etcd.put(action.key, action.value)
        return SetConfigActionResult()

    async def delete_config(self, action: DeleteConfigAction) -> DeleteConfigActionResult:
        """Delete raw etcd key-value."""
        if action.prefix:
            await self._etcd.delete_prefix(action.key)
        else:
            await self._etcd.delete(action.key)
        return DeleteConfigActionResult()
